"""
    Homonim: Correction of aerial and satellite imagery to surface reflectance
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import concurrent.futures
import logging
import pathlib
import threading
from typing import Dict, List, Tuple

import numpy as np
from rasterio.warp import Resampling
from tabulate import tabulate
from tqdm import tqdm

from homonim import utils
from homonim.enums import ProcCrs
from homonim.raster_pair import RasterPairReader, BlockPair

logger = logging.getLogger(__name__)


class RasterCompare(RasterPairReader):

    def __init__(self, src_filename: pathlib.Path, ref_filename: pathlib.Path, proc_crs: ProcCrs = ProcCrs.auto):
        """
        Class to statistically compare source and reference images.

        Parameters
        ----------
        src_filename: str, pathlib.Path
            Path to the source image file.  Can be any raw, corrected, etc. multi-spectral image.
        ref_filename: str, pathlib.Path
            Path to the reference image file.  The extents of this image should cover the source with at least a 2
            pixel border.  The reference image should have at least as many bands as the source, and the
            ordering of the source and reference bands should match.
        proc_crs: homonim.enums.ProcCrs, optional
            :class:`~homonim.enums.ProcCrs` instance specifying which of the source/reference image spaces will be
            used for comparison.  In most cases, it can be left as the default of
            :attr:`~homonim.enums.ProcCrs.auto`,  where it will be resolved to the lowest resolution of the source and
            reference image CRS's.
        """
        RasterPairReader.__init__(self, src_filename, ref_filename, proc_crs=proc_crs)
        self._lock = threading.Lock()

    schema = dict(
        r2=dict(abbrev='r\N{SUPERSCRIPT TWO}', description='Pearson\'s correlation coefficient squared', ),
        rmse=dict(abbrev='RMSE', description='Root Mean Square Error', ),
        rrmse=dict(abbrev='rRMSE', description='Relative RMSE (RMSE/mean(ref))', ),
        n=dict(abbrev='N', description='Number of pixels', )
    )  # yapf: disable
    """ Dictionary describing the statistics returned by :attr:`RasterCompare.compare`. """

    @property
    def schema_table(self) -> str:
        """ Table string describing statistics returned by :attr:`RasterCompare.compare`. """
        headers = {key: key.upper() for key in list(self.schema.values())[0].keys()}
        return tabulate(self.schema.values(), headers=headers, tablefmt=utils.table_format)

    @staticmethod
    def create_config(
        threads: int = 0, max_block_mem: float = 512, downsampling: Resampling = Resampling.average,
        upsampling: Resampling = Resampling.cubic_spline,
    ) -> Dict:
        """
        Utility method to create a RasterCompare configuration dictionary that can be passed to
        :meth:`RasterCompare.process`.  Without arguments, the default configuration values are returned.

        Parameters
        ----------
        threads: int, optional
            Number of image blocks to process concurrently.  A maximum of the number of processors on your
            system is allowed.  Increasing this number will increase the memory required for processing.
            0 = use all processors.
        max_block_mem: float, optional
            Maximum size of an image block in megabytes. Note that the total memory consumed by a thread is
            proportional to, but a number of times larger than this number.
        downsampling: rasterio.enums.Resampling, optional
            Resampling method to use when downsampling. See the `rasterio docs
            <https://rasterio.readthedocs.io/en/latest/api/rasterio.enums.html#rasterio.enums.Resampling>`_ for
            available options.
        upsampling: rasterio.enums.Resampling, optional
            Resampling method to use when upsampling.  See the `rasterio docs
            <https://rasterio.readthedocs.io/en/latest/api/rasterio.enums.html#rasterio.enums.Resampling>`_ for
            available options.

        Returns
        -------
        dict
            Configuration dictionary.
        """
        return dict(
            threads=utils.validate_threads(threads), max_block_mem=max_block_mem, downsampling=downsampling,
            upsampling=upsampling,
        )

    def _get_resampling(self, from_res: Tuple[float, float], to_res: Tuple[float, float], **kwargs) -> Resampling:
        """
        Utility method to return the resampling method for re-projecting from resolution `from_res` to resolution
        `to_res`.
        """
        config = self.create_config(**kwargs)
        return config['downsampling'] if np.prod(np.abs(from_res)) <= np.prod(np.abs(to_res)) else config['upsampling']

    def _get_image_stats(self, image_sums: List[Dict]) -> List[Dict]:
        """ Return the band comparison statistics, given src, ref, src**2 etc band sums. """

        def get_band_stats(
            src_sum: float = 0, ref_sum: float = 0, src2_sum: float = 0, ref2_sum: float = 0, src_ref_sum: float = 0,
            res2_sum: float = 0, mask_sum: float = 0
        ) -> Dict:
            """ Return the comparison statistics for a band, given the source, reference etc. sums. """
            # find PCC using the 3rd equation down at
            # https://en.wikipedia.org/wiki/Pearson_correlation_coefficient#For_a_sample
            src_mean = src_sum / mask_sum
            ref_mean = ref_sum / mask_sum
            pcc_num = src_ref_sum - (mask_sum * src_mean * ref_mean)
            pcc_den = (
                np.sqrt(src2_sum - (mask_sum * (src_mean ** 2))) * np.sqrt(ref2_sum - (mask_sum * (ref_mean ** 2)))
            )
            pcc = pcc_num / pcc_den

            # find RMSE and rRMSE
            rmse = np.sqrt(res2_sum / mask_sum)
            rrmse = rmse / ref_mean
            return dict(r2=pcc ** 2, rmse=rmse, rrmse=rrmse, n=int(mask_sum))

        image_stats = []
        sum_over_bands = {}
        for band_i, band_sum_dict in enumerate(image_sums):
            band_stats = get_band_stats(**band_sum_dict)
            band_desc = (
                self.ref_im.descriptions[self.ref_bands[band_i] - 1] or
                self.src_im.descriptions[self.src_bands[band_i] - 1] or
                f'Band {band_i + 1}'
            )  # yapf: disable
            image_stats.append(dict(band=band_desc, **band_stats))
            sum_over_bands = {k: sum_over_bands.get(k, 0) + v for k, v in band_stats.items()}

        # find mean of each statistic over the bands, retaining int types
        mean_stats = {
            k: int(v / len(image_sums)) if isinstance(v, int) else (v / len(image_sums))
            for k, v in sum_over_bands.items()
        }  # yapf: disable
        # add the means to the list of bands
        image_stats.append(dict(band='Mean', **mean_stats))
        return image_stats

    def stats_table(self, stats_list: List[Dict]):
        """
        Create a table string from the provided comparison statistics.

        Parameters
        ----------
        stats_list: list of dict
            Comparison statistics to tabulate, as returned by :meth:`RasterCompare.compare`.

        Returns
        -------
        str
            Table string.
        """
        headers = {
            k: self.schema[k]['abbrev'] if k in self.schema else str.capitalize(k)
            for k in list(stats_list[0].keys())
        }  # yapf: disable
        return tabulate(stats_list, headers=headers, floatfmt='.3f', stralign='right', tablefmt=utils.table_format)

    def compare(self, **kwargs) -> List[Dict]:
        """
        Statistically compare source and reference images.

        To improve speed and reduce memory usage, images are divided into blocks for concurrent processing.

        Parameters
        ----------
        kwargs
            Optional configuration settings.  See :meth:`RasterCompare.create_config` for possible arguments and their
            default values.

        Returns
        -------
        list of dict
            List of dicts for each band, representing the comparison results.
        """
        self._assert_open()
        config = self.create_config(**kwargs)

        def get_block_sums(block_pair: BlockPair):
            """ Thread-safe function to find the source/reference sums for a block.  """
            src_ra, ref_ra = self.read(block_pair)  # read src and ref blocks

            # re-project so that both source and reference are in proc_crs
            if self.proc_crs == ProcCrs.ref:
                resampling = self._get_resampling(src_ra.res, ref_ra.res, **kwargs)
                src_ra = src_ra.reproject(**ref_ra.proj_profile, resampling=resampling)
            else:
                resampling = self._get_resampling(ref_ra.res, src_ra.res, **kwargs)
                ref_ra = ref_ra.reproject(**src_ra.proj_profile, resampling=resampling)

            # mask invalid pixels so they don't contribute to sums
            src_array = src_ra.array
            ref_array = ref_ra.array
            mask = ref_ra.mask & src_ra.mask
            src_array[~mask] = 0
            ref_array[~mask] = 0
            # find the required sums and return
            sums_dict = dict(
                src_sum=src_array.sum(), ref_sum=ref_array.sum(), src2_sum=(src_array ** 2).sum(),
                ref2_sum=(ref_array ** 2).sum(), src_ref_sum=(src_array * ref_array).sum(),
                res2_sum=((ref_array - src_array) ** 2).sum(), mask_sum=mask.sum()
            )
            return sums_dict, block_pair

        with concurrent.futures.ThreadPoolExecutor(max_workers=config['threads']) as executor:
            # read and sum image blocks in threads
            futures = [
                executor.submit(get_block_sums, block_pair)
                for block_pair in self.block_pairs(max_block_mem=config['max_block_mem'])
            ]  # yapf: disable

            # wait for threads
            image_sums = [{} for _ in self.src_bands]
            bar_format = '{l_bar}{bar}|{n_fmt}/{total_fmt} blocks [{elapsed}<{remaining}]'
            for future in tqdm(
                concurrent.futures.as_completed(futures), bar_format=bar_format, total=len(futures), dynamic_ncols=True,
            ):  # yapf: disable
                # get block sums and accumulate over the image
                block_sums_dict, block_pair = future.result()
                image_sums[block_pair.band_i] = (
                    {k: image_sums[block_pair.band_i].get(k, 0) + v for k, v in block_sums_dict.items()}
                )  # yapf: disable

        # find the comparison statistics from the accumulated block sums, and return
        return self._get_image_stats(image_sums)



"""
    Homonim: Correction of aerial and satellite imagery to surface reflectance
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import logging
import multiprocessing
import pathlib
from typing import Tuple, Dict, List, Union, Optional

import numpy
import numpy as np
import rasterio as rio
from rasterio import Affine, windows
from rasterio.crs import CRS
from rasterio.enums import MaskFlags
from rasterio.transform import TransformMethodsMixin
from rasterio.warp import reproject, Resampling
from rasterio.windows import Window, WindowMethodsMixin

from homonim import utils
from homonim.errors import ImageProfileError, ImageFormatError

logger = logging.getLogger(__name__)


class RasterArray(TransformMethodsMixin, WindowMethodsMixin):
    """
    A class for encapsulating a masked, geo-referenced numpy array.
    Provides methods for re-projection, and reading/writing from/to rasterio datasets.
    """
    default_nodata = float('nan')  # default internal nodata value
    default_dtype = 'float32'  # default internal data type

    def __init__(
        self, array: np.ndarray, crs: CRS, transform: Affine, nodata: Optional[float] = default_nodata,
        window: Optional[Window] = None
    ):
        """
        Construct a RasterArray.

        Parameters
        ----------
        array: numpy.ndarray
            2 or 3D array of image data, if 3D, bands are along the first dimension.
        crs: rasterio.crs.CRS
            ``array`` CRS.
        transform: rasterio.transform.Affine
            ``array`` geo-transform.
        nodata: optional
            A number or nan, specifying the nodata value to use for masking the array.
        window: rasterio.windows.Window, optional
            Optional window into the transform specifying the array region.
        """

        if (array.ndim < 2) or (array.ndim > 3):
            raise ValueError('`array` must be have 2 or 3 dimensions with bands along the first dimension')
        self._array = array

        if window is not None and (window.height, window.width) != array.shape[-2:]:
            raise ValueError('`window` and `array` width and height must match')

        if isinstance(crs, CRS):
            self._crs = crs
        else:
            raise TypeError('`crs` must be an instance of rasterio.crs.CRS')

        if isinstance(transform, Affine):
            if window is not None:
                self._transform = windows.transform(window, transform)
            else:
                self._transform = transform
        else:
            raise TypeError("`transform` must be an instance of rasterio.transform.Affine")

        self._nodata = nodata
        self._nodata_mask = None

    @classmethod
    def from_profile(cls, array: Optional[np.ndarray], profile: Dict, window: Optional[Window] = None) -> 'RasterArray':
        """
        Construct a RasterArray from an array of image data and a profile dictionary.

        Parameters
        ----------
        array: numpy.ndarray, None
            2 or 3D array of image data, if 3D, bands are along the first dimension.
            Can be None, in which case a nodata array is created with the `width`, `height`, `count`, `dtype` and
            `nodata` fields in ``profile``.
        profile: dict
            Configuration dictionary with items specifying the `crs`, `transform` and `nodata` values
            (as used by rasterio datasets).  If ``array`` is None, this dict should contain the additional fields
            to create the array.
        window: rasterio.windows.Window, optional
            Optional window into ``profile['transform']``, specifying the array region.

        Returns
        -------
        RasterArray
            Constructed RasterArray.
        """
        if not {'crs', 'transform', 'nodata'} <= set(profile):
            raise ImageProfileError("'profile' should include 'crs', 'transform' and 'nodata' keys")

        if array is None:  # create array filled with nodata
            if not {'width', 'height', 'count', 'dtype'} <= set(profile):
                raise ImageProfileError("'profile' should include 'width', 'height', 'count' and 'dtype' keys")
            array_shape = (profile['count'], profile['height'], profile['width'])
            array = np.full(array_shape, fill_value=profile['nodata'], dtype=profile['dtype'])

        return cls(array, profile['crs'], profile['transform'], nodata=profile['nodata'], window=window)

    @classmethod
    def from_rio_dataset(
        cls, rio_dataset: rio.DatasetReader, indexes: Optional[Union[int, List[int]]] = None,
        window: Optional[Window] = None, **kwargs
    ) -> 'RasterArray':
        """
        Construct a RasterArray by reading from a rasterio dataset.

        Implements 'boundless' reads internally which is faster than using rasterio's boundless=True option.

        Parameters
        ----------
        rio_dataset: rasterio.DatasetReader
            Rasterio dataset to be read from.
        indexes: int, list[int], optional
            1-based index or list of indexes of the bands to be read from ``rio_dataset``.
            The default is to read all the ``rio_dataset`` bands.
        window: rasterio.windows.Window, optional
            Optional window into ``rio_dataset`` to be read from.
            This can be a `boundless` window i.e. a window that extends beyond the bounds of ``rio_dataset``,
            in which case the :attr:`~RasterArray.array` will be filled with nodata outside the ``rio_dataset``
            bounds.
        kwargs: dict, optional
            Additional arguments to be passed to the dataset's read() method.

        Returns
        -------
        RasterArray
            Constructed RasterArray.
        """
        # form a list of indexes
        if indexes is None:
            index_list = utils.get_nonalpha_bands(rio_dataset)
        else:
            index_list = [indexes] if np.isscalar(indexes) else indexes

        if window is None:
            # window of the full dataset extent
            window = Window(col_off=0, row_off=0, width=rio_dataset.width, height=rio_dataset.height)

        # check bands if bands have masks (i.e. internal/side-car mask or alpha channel), as opposed to nodata value
        is_masked = any([MaskFlags.per_dataset in rio_dataset.mask_flag_enums[bi - 1] for bi in index_list])

        # use the dataset's nodata value if it 'unmasked', and has one, otherwise revert to default
        nodata = cls.default_nodata if (is_masked or rio_dataset.nodata is None) else rio_dataset.nodata

        # construct an array of nodata matching the (possibly boundless) window dimension
        bounded_window, bounded_slices = cls.bounded_window_slices(rio_dataset, window)
        if len(index_list) > 1:
            array = np.full((len(index_list), window.height, window.width), fill_value=nodata, dtype=cls.default_dtype)
            bounded_array = array[(slice(array.shape[0]), *bounded_slices)]  # a bounded view into array
            rio_dataset.read(
                out=bounded_array, indexes=index_list, window=bounded_window, out_dtype=cls.default_dtype, **kwargs
            )
        else:
            array = np.full((window.height, window.width), fill_value=nodata, dtype=cls.default_dtype)
            bounded_array = array[bounded_slices]  # a bounded view into array
            rio_dataset.read(
                out=bounded_array, indexes=index_list[0], window=bounded_window, out_dtype=cls.default_dtype, **kwargs
            )

        # read into the bounded section of the array
        if is_masked:
            # read the mask from dataset and apply it to the array
            bounded_mask = rio_dataset.dataset_mask(window=bounded_window).astype('bool', copy=False)
            if bounded_array.ndim == 2:
                bounded_array[~bounded_mask] = nodata
            else:
                bounded_array[:, ~bounded_mask] = nodata

        return cls(array, rio_dataset.crs, rio_dataset.transform, nodata=nodata, window=window)

    @staticmethod
    def bounded_window_slices(
        rio_dataset: Union[rio.DatasetReader, rio.io.DatasetWriter], window: Window
    ) -> Tuple[Window, Tuple[slice, slice]]:  # yapf: disable
        """ Bounded array slices and dataset window from dataset and boundless window. """

        # find window UL and BR corners and crop to rio_dataset bounds
        win_ul = np.array((window.row_off, window.col_off))
        win_br = win_ul + np.array((window.height, window.width))
        bounded_ul = np.fmax(win_ul, (0, 0))
        bounded_br = np.fmin(win_br, rio_dataset.shape)

        # create bounded window and slices from bounded corners
        bounded_window = Window.from_slices((bounded_ul[0], bounded_br[0]), (bounded_ul[1], bounded_br[1]))
        bounded_start = bounded_ul - win_ul
        bounded_stop = bounded_start + (bounded_br - bounded_ul)
        bounded_slices = (
            slice(bounded_start[0], bounded_stop[0], None),
            slice(bounded_start[1], bounded_stop[1], None)
        )  # yapf: disable
        return bounded_window, bounded_slices

    @property
    def array(self) -> numpy.ndarray:
        """ 2 or 3D array of image data, if 3D, bands are along the first dimension. """
        return self._array

    @array.setter
    def array(self, value: numpy.ndarray):
        if np.all(value.shape[-2:] == self._array.shape[-2:]):
            self._array = value
        else:
            raise ValueError("'value' and 'array' shapes must match")

    @property
    def crs(self) -> CRS:
        """ Coordinate reference system. """
        return self._crs

    @property
    def width(self) -> int:
        """ Array width in pixels. """
        return self.shape[-1]

    @property
    def height(self) -> int:
        """ Array height in pixels. """
        return self.shape[-2]

    @property
    def shape(self) -> Tuple[int, int]:
        """ Array shape (height, width) in pixels. """
        return tuple(self._array.shape[-2:])

    @property
    def count(self) -> int:
        """ Number of bands. """
        return self._array.shape[0] if self.array.ndim == 3 else 1

    @property
    def dtype(self) -> str:
        """ Internal data type of the array. """
        return self._array.dtype.name

    @property
    def transform(self) -> Affine:
        """ Affine geo-transform describing the location and orientation of the array in the CRS. """
        return self._transform

    @property
    def res(self) -> Tuple[float, float]:
        """ Array (x, y) resolution (m). """
        return self._transform.a, -self._transform.e

    @property
    def bounds(self) -> Tuple[float, ...]:
        """ (left, bottom, right, top) co-ordinates of the array extent. """
        return windows.bounds(windows.Window(0, 0, self.width, self.height), self._transform)

    @property
    def profile(self) -> Dict:
        """ RasterArray properties formatted as a dictionary, compatible with rasterio. """
        return dict(
            crs=self._crs, transform=self._transform, nodata=self._nodata, count=self.count, width=self.width,
            height=self.height, dtype=self.dtype
        )

    @property
    def proj_profile(self) -> Dict:
        """
        RasterArray properties relevant to re-projection (i.e. `crs`, `transform` and `shape`) formatted as a
        dictionary.
        Useful for expanding to keyword arguments to :meth:`reproject`.
        """
        return dict(crs=self._crs, transform=self._transform, shape=self.shape)

    @property
    def mask(self) -> numpy.ndarray:
        """ 2D boolean mask corresponding to valid pixels in the array. """
        if self._nodata is None:
            return np.full(self._array.shape[-2:], True)
        mask = ~utils.nan_equals(self._array, self._nodata)
        if self._array.ndim > 2:
            mask = np.any(mask, axis=0)
        return mask

    @mask.setter
    def mask(self, value: numpy.ndarray):
        if self._array.ndim == 2:
            self._array[~value] = self._nodata
        else:
            self._array[:, ~value] = self._nodata

    @property
    def mask_ra(self) -> 'RasterArray':
        """
        RasterArray containing the 2D mask as `uint8` data type, and with nodata=None.  Useful for re-projecting
        the mask.
        """
        mask = self.mask.astype('uint8', copy=False)
        return RasterArray(mask, crs=self._crs, transform=self._transform, nodata=None)

    @property
    def nodata(self) -> float:
        """ Nodata value. """
        return self._nodata

    @nodata.setter
    def nodata(self, value: float):
        if value is None or self._nodata is None:
            # if new nodata value is None, remove the current mask
            # if current nodata is None, there is no mask, so just set the new nodata value and return
            self._nodata = value
        elif not (utils.nan_equals(value, self._nodata)):
            # if the new nodata value is different to the current nodata,
            # set the mask area in array to the new nodata value and return
            nodata_mask = ~self.mask
            if self._array.ndim == 3:
                self._array[:, nodata_mask] = value
            else:
                self._array[nodata_mask] = value
            self._nodata = value

    def copy(self) -> 'RasterArray':
        """ Create a deep copy of the RasterArray. """
        return RasterArray.from_profile(self._array.copy(), self.profile)

    def slice_to_bounds(self, *bounds) -> 'RasterArray':
        """
        Create a new RasterArray representing a rectangular subregion of this RasterArray.
        Note that the created RasterArray is a view into the current array, not a copy.

        Parameters
        ----------
        bounds: Tuple
            Co-ordinate bounds to slice the new array to (left, bottom, right, top), in the current :attr:`crs`.

        Returns
        -------
        RasterArray
            Sliced RasterArray.
        """
        window = self.window(*bounds)
        window = utils.round_window_to_grid(window)
        ul = np.array((window.row_off, window.col_off))
        shape = np.array((window.height, window.width))
        if np.any(ul < 0) or np.any(shape > self._array.shape[-2:]):
            raise ValueError(
                f'The provided bounds ({bounds}) lie outside the extent of the RasterArray ({self.bounds})'
            )

        if self._array.ndim == 2:
            array = self._array[window.toslices()]
        else:
            array = self._array[(slice(self._array.shape[0]), *window.toslices())]

        return RasterArray(array, self._crs, self.window_transform(window), nodata=self._nodata)

    def to_rio_dataset(
        self, rio_dataset: rio.io.DatasetWriter, indexes: Optional[Union[int, List[int]]] = None,
        window: Optional[Window] = None, **kwargs
    ):
        """
        Write the RasterArray into a rasterio dataset.

        Note that typically, the dataset bounds would encompass the RasterArray bounds.

        Parameters
        ----------
        rio_dataset: rasterio.io.DatasetWriter
            An open rasterio dataset into which to write the RasterArray.
            The dataset CRS must match that of the RasterArray.
        indexes: int, list[int], optional
            1-based index or list of indexes of the bands to be written in ``rio_dataset``.
            It should contain the same number of items as there are RasterArray bands.
            The default is to write into the first `count` non-alpha bands of the dataset, where `count` is the number
            of RasterArray bands.
        window: rasterio.windows.Window, optional
            Window defining the region in the dataset to write the RasterArray to, and how to crop the
            RasterArray, if necessary.  If it is a `boundless` window i.e. extended beyond the bounds of the
            dataset, it is cropped to fit the bounds of the dataset. The RasterArray is cropped to fit the bounds
            of the window in the dataset.
            The default is to write the full extent of RasterArray into the corresponding region in the dataset.
        kwargs: optional
            Arguments to pass through the dataset's write() method.
        """
        if not np.all(np.abs(self.res) == np.abs(rio_dataset.res)):
            raise ImageFormatError(
                f'The dataset resolution does not match that of the RasterArray. '
                f'Dataset res: {rio_dataset.res}, RasterArray res: {self.res}'
            )
        # TODO : raise an issue with rasterio about the speed of crs comparison.  comparing the _crs attr is faster.
        if self.crs != rio_dataset.crs:
            raise ImageFormatError(
                f'The dataset CRS does not match that of the RasterArray. '
                f'Dataset CRS: {rio_dataset.crs.to_proj4()}, RasterArray CRS: {self.crs.to_proj4()}'
            )

        if indexes is None:
            indexes = utils.get_nonalpha_bands(rio_dataset)
            indexes = indexes if len(indexes) > 1 else indexes[0]

        if np.any((np.array(indexes) < 1) | (np.array(indexes) > rio_dataset.count)):
            error_indexes = np.array(indexes)[np.array(indexes) > rio_dataset.count]
            raise ValueError(f'Band index(es) {error_indexes} are out of the valid range (1..{rio_dataset.count})')

        if (not np.isscalar(indexes)) and (len(indexes) > self.count):
            raise ValueError(
                f'The length of indexes ({len(indexes)}) exceeds the number of bands in the '
                f'RasterArray ({self.count})'
            )

        if window is None:
            # a window defining the region in the dataset corresponding to the RasterArray extents
            window = rio_dataset.window(*self.bounds)

        # crop the window to dataset bounds
        window, _ = self.bounded_window_slices(rio_dataset, window)
        # crop the RasterArray to match the bounds of the dataset window
        bounded_ra = self.slice_to_bounds(*rio_dataset.window_bounds(window))

        if np.any(bounded_ra.shape != np.array((window.height, window.width))):
            raise ValueError(
                f'The bounds of the dataset / window ({rio_dataset.window_bounds(window)}) lie outside the '
                f'bounds of the RasterArray ({bounded_ra.bounds})'
            )

        rio_dataset.write(bounded_ra.array, window=window, indexes=indexes, **kwargs)

    def to_file(self, filename: Union[str, pathlib.Path], driver: str = 'GTiff', **kwargs):
        """
        Write the RasterArray to an image file.

        Parameters
        ----------
        filename: str, pathlib.Path
            Name of the file to create.
        driver: str, optional
            Valid rasterio short format driver name - See the `GDAL docs <https://gdal.org/drivers/raster/index.html>`_
            for available options.
        kwargs: dict, optional
            Driver specific creation options e.g. ``compression='deflate'`` for a GeoTiff.
            See the `GDAL docs <https://gdal.org/drivers/raster/index.html>`_ for available keys and values.
        """
        with rio.Env(GDAL_NUM_THREADS='ALL_CPUs'):
            with rio.open(filename, 'w', driver=driver, **self.profile, **kwargs) as out_im:
                out_im.write(self._array, indexes=range(1, self.count + 1) if self.count > 1 else 1)

    def reproject(
        self, crs: Optional[CRS] = None, transform: Optional[Affine] = None, shape: Optional[Tuple[int, int]] = None,
        nodata: float = default_nodata, dtype: str = default_dtype, resampling: Resampling = Resampling.lanczos,
        **kwargs
    ):
        """
        Re-project the RasterArray.

        Parameters
        ----------
        crs: rasterio.crs.CRS, optional
             CRS to project into.  The default is to use the CRS of this RasterArray.
        transform: rasterio.transform.Affine, optional
            Geo-transform to project into.  If ``transform`` is specified, ``shape`` is also required.
            The default is to use the transform of this RasterArray.
        shape: tuple, optional
            (rows, columns) size of the destination array. The default is to use the shape of this RasterArray.
        nodata: float, int, optional
            nodata value of the destination array.
        dtype: type, str, optional
            Internal data type of destination array.
        resampling: rasterio.enums.Resampling, optional
            Resampling method to use.
        kwargs: dict, optional
            Arguments to passed through the rasterio's reproject() function.

        Returns
        -------
        RasterArray
            Reprojected RasterArray.
        """

        if transform is not None and shape is None:
            raise ValueError('If `transform` is specified, `shape` must also be specified')

        if isinstance(resampling, str):
            resampling = Resampling[resampling]

        crs = crs or self.crs
        shape = shape or self.shape
        dtype = dtype or self.dtype

        if self.array.ndim > 2:
            _dst_array = np.zeros((self._array.shape[0], *shape), dtype=dtype)
        else:
            _dst_array = np.zeros(shape, dtype=dtype)

        _, _dst_transform = reproject(
            self._array, destination=_dst_array, src_crs=self._crs, src_transform=self._transform,
            src_nodata=self._nodata, dst_crs=crs, dst_transform=transform, dst_nodata=nodata,
            num_threads=multiprocessing.cpu_count(), resampling=resampling, **kwargs
        )
        return RasterArray(_dst_array, crs=crs, transform=_dst_transform, nodata=nodata)


##



"""
    Homonim: Correction of aerial and satellite imagery to surface reflectance
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import os
import pathlib

from homonim.compare import RasterCompare
from homonim.enums import Model, ProcCrs
from homonim.fuse import RasterFuse
from homonim.kernel_model import KernelModel
from homonim.stats import ParamStats

if '__file__' in globals():
    root_path = pathlib.Path(__file__).absolute().parents[1]
else:
    root_path = pathlib.Path(os.getcwd())



"""
    Homonim: Correction of aerial and satellite imagery to surface reflectance
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import json
import logging
import pathlib
import re
import sys
from timeit import default_timer as timer
from typing import Tuple, Dict

import click
import cloup
import numpy as np
import rasterio as rio
import yaml
from click.core import ParameterSource
from rasterio.warp import SUPPORTED_RESAMPLING

from homonim import utils, version, RasterFuse, RasterCompare, ParamStats, ProcCrs, Model
from homonim.errors import ImageFormatError
from homonim.raster_array import RasterArray

logger = logging.getLogger(__name__)


class PlainInfoFormatter(logging.Formatter):
    """ logging formatter to format INFO logs without the module name etc. prefix. """

    def format(self, record: logging.LogRecord):
        if record.levelno == logging.INFO:
            self._style._fmt = '%(message)s'
        else:
            self._style._fmt = '%(levelname)s:%(name)s: %(message)s'
        return super().format(record)


class HomonimCommand(cloup.Command):
    """ cloup.Command subclass for formatting help with RST markup. """

    def get_help(self, ctx: click.Context):
        """ Strip some RST markup from the help text for CLI display.  Will not work with grid tables. """

        # Note that this can't easily be done in __init__, as each sub-command's __init__ gets called,
        # which ends up re-assigning self.wrap_text to reformat_text
        if not hasattr(self, 'wrap_text'):
            self.wrap_text = cloup.formatting._formatter.wrap_text
        sub_strings = {
            '\b\n': '\n\b',  # convert from RST friendly to click literal (unwrapped) block marker
            r'\| ': '',  # strip RST literal (unwrapped) marker in e.g. tables and bullet lists
            r'\n\.\. _.*:\n': '',  # strip RST ref directive '\n.. <name>:\n'
            '`(.*?) <(.*?)>`_': r'\g<1>',  # convert from RST cross-ref '`<name> <<link>>`_' to 'name'
            '::': ':',  # convert from RST '::' to ':'
            '``(.*?)``': r'\g<1>',  # convert from RST '``literal``' to 'literal'
            ':option:`(.*?)( <.*?>)?`': r'\g<1>',  # convert ':option:`--name <group-command --name>`' to '--name'
            ':option:`(.*?)`': r'\g<1>',  # convert ':option:`--name`' to '--name'
        }  # yapf: disable

        def reformat_text(text: str, width: int, **kwargs):
            for sub_key, sub_value in sub_strings.items():
                text = re.sub(sub_key, sub_value, text, flags=re.DOTALL)
            return self.wrap_text(text, width, **kwargs)

        cloup.formatting._formatter.wrap_text = reformat_text
        return cloup.Command.get_help(self, ctx)


class FuseCommand(HomonimCommand):
    """ click.Command subclass for setting fuse command parameters from a yaml config file. """

    def invoke(self, ctx: click.Context):
        """ Merge config file with command line and default parameter values.  """
        # adapted from https://stackoverflow.com/questions/46358797/python-click-supply-arguments-and-options-from-a
        # -configuration-file/46391887
        config_file = ctx.params['conf']
        if config_file is not None:
            # read the config file into a dict
            with open(config_file) as f:
                config_dict = yaml.safe_load(f)

            for conf_key, conf_value in config_dict.items():
                if conf_key not in ctx.params:
                    raise click.BadParameter(f'Unknown config file parameter "{conf_key}"', ctx=ctx, param_hint='conf')
                else:
                    param_src = ctx.get_parameter_source(conf_key)
                    # overwrite parameters not specified on command line with config file values
                    if ctx.params[conf_key] is None or param_src == ParameterSource.DEFAULT:
                        ctx.params[conf_key] = conf_value
                        ctx.set_parameter_source(conf_key, ParameterSource.COMMANDLINE)

        # Set the default creation_options if no other driver or creation_options have been specified.
        # This can't be done in a callback as it depends on --driver.
        if (ctx.get_parameter_source('driver') == ParameterSource.DEFAULT and
                ctx.get_parameter_source('creation_options') == ParameterSource.DEFAULT):
            ctx.params['creation_options'] = RasterFuse.create_out_profile()['creation_options']

        return click.Command.invoke(self, ctx)


def _update_existing_keys(default_dict: Dict, **kwargs) -> Dict:
    """ Update values in a dict with args from matching keys in **kwargs. """
    return {k: kwargs.get(k, v) for k, v in default_dict.items()}


def _configure_logging(verbosity: int):
    """ Configure python logging level."""
    # adapted from rasterio https://github.com/rasterio/rasterio
    log_level = max(10, 20 - 10 * verbosity)

    # apply config to package logger, rather than root logger
    pkg_logger = logging.getLogger(__package__)
    formatter = PlainInfoFormatter()
    handler = logging.StreamHandler(sys.stderr)
    handler.setFormatter(formatter)
    pkg_logger.addHandler(handler)
    pkg_logger.setLevel(log_level)
    logging.captureWarnings(True)


def _threads_cb(ctx: click.Context, param: click.Option, value):
    """ click callback to validate --threads. """
    try:
        threads = utils.validate_threads(value)
    except Exception as ex:
        raise click.BadParameter(str(ex))
    return threads


def _nodata_cb(ctx: click.Context, param: click.Option, value):
    """ click callback to convert --nodata value to None, nan or float. """
    # adapted from rasterio https://github.com/rasterio/rasterio
    if value is None or value.lower() in ['null', 'nil', 'none', 'nada']:
        return None
    else:
        # check value is a number and can be cast to output dtype
        try:
            value = float(value.lower())
            if not rio.dtypes.can_cast_dtype(value, ctx.params['dtype']):
                raise click.BadParameter(
                    f'{value} cannot be cast to the output image data type {ctx.params["dtype"]}', param=param,
                    param_hint='nodata'
                )
        except (TypeError, ValueError):
            raise click.BadParameter(f'{value} is not a number', param=param, param_hint='nodata')

        return value


def _compare_cb(ctx: click.Context, param: click.Option, value):
    """ click callback to check --compare path exists if specified.  """
    if value and str(value) != 'ref' and not pathlib.Path(value).exists():
        raise click.BadParameter(f'Comparison image does not exist: {value}')
    return value


def _creation_options_cb(ctx: click.Context, param: click.Option, value):
    """
    click callback to validate and parse multiple creation options (e.g. `-co KEY1=VAL1 -co KEY2=VAL2).
    Note: `==VAL` breaks this as `str.split('=', 1)` is used.
    """
    # adapted from rasterio https://github.com/rasterio/rasterio
    if not value:
        return {}
    else:
        out = {}
        for pair in value:
            if '=' not in pair:
                raise click.BadParameter('Invalid syntax for KEY=VAL arg: {}'.format(pair))
            else:
                k, v = pair.split('=', 1)
                k = k.lower()
                v = v.lower()
                out[k] = None if v.lower() in ['none', 'null', 'nil', 'nada'] else yaml.safe_load(v)
        return out


def _param_file_cb(ctx: click.Context, param: click.Argument, value):
    """ click callback to validate parameter image file(s). """
    for filename in value:
        filename = pathlib.Path(filename)
        try:
            utils.validate_param_image(filename)
        except (FileNotFoundError, ImageFormatError):
            raise click.BadParameter(f'{filename.name} is not a valid parameter image.', param=param)
    return value


# define click options and arguments common to more than one command
# use cloup's argument to auto print argument help on command line
ref_file_arg = cloup.argument(
    'ref-file', nargs=1, metavar='REFERENCE', type=click.Path(exists=True, dir_okay=False, path_type=pathlib.Path),
    help='Path to a reference image.'
)
threads_option = click.option(
    '-t', '--threads', type=click.INT, default=RasterFuse.create_block_config()['threads'], show_default=True,
    callback=_threads_cb, help=f'Number of image blocks to process concurrently (0 = use all processors).'
)
max_block_mem_option = click.option(
    '-mbm', '--max-block-mem', type=click.FLOAT, default=RasterFuse.create_block_config()['max_block_mem'],
    show_default=True, help='Maximum image block size in megabytes (0 = block corresponds to a whole band).'
)
downsampling_option = click.option(
    '-ds', '--downsampling', type=click.Choice([r.name for r in rio.warp.SUPPORTED_RESAMPLING]),
    default=RasterFuse.create_model_config()['downsampling'].name, show_default=True,
    help='Resampling method for re-projecting from high to low resolution.  See the `rasterio docs '
    '<https://rasterio.readthedocs.io/en/latest/api/rasterio.enums.html#rasterio.enums.Resampling>`_ for '
    'details.'
)
upsampling_option = click.option(
    '-us', '--upsampling', type=click.Choice([r.name for r in rio.warp.SUPPORTED_RESAMPLING]),
    default=RasterFuse.create_model_config()['upsampling'].name, show_default=True,
    help='Resampling method for re-projecting from low to high resolution.  See the `rasterio docs '
    '<https://rasterio.readthedocs.io/en/latest/api/rasterio.enums.html#rasterio.enums.Resampling>`_ for '
    'details.'
)
output_option = click.option(
    '-op', '--output', type=click.Path(exists=False, dir_okay=False, writable=True, path_type=pathlib.Path),
    help='Write results to this json file.'
)
""" cloup context settings to print help in 'linear' layout with heading/option emphasis. """
context_settings = cloup.Context.settings(
    formatter_settings=cloup.HelpFormatter.settings(
        col2_min_width=np.inf,
        theme=cloup.HelpTheme(
            invoked_command=cloup.Style(fg='bright_white', bold=True),
            heading=cloup.Style(fg='bright_white', bold=True),
            col1=cloup.Style(fg='bright_white'),
        )
    )
)  # yapf: disable


# define the click CLI
@cloup.group(context_settings=context_settings)
@click.option('--verbose', '-v', count=True, help='Increase verbosity.')
@click.option('--quiet', '-q', count=True, help='Decrease verbosity.')
@click.version_option(version=version.__version__, message='%(version)s')
def cli(verbose: int, quiet: int):
    """ Surface reflectance correction and support utilities. """
    verbosity = verbose - quiet
    _configure_logging(verbosity)


# fuse command
@cloup.command(cls=FuseCommand)
# standard options
@cloup.argument(
    'src-file', nargs=-1, metavar='SOURCE...', type=click.Path(exists=True, dir_okay=False, path_type=pathlib.Path),
    help='Path(s) to source image(s) to be corrected.'
)
@ref_file_arg
@cloup.option_group(
    "Standard options",
    # note: either use click.option(...), or cloup.option(..., help=inspect.cleandoc(...)) for RST help strings,
    # if cloup's mutually exclusive etc. functionality is needed, it should be the latter.
    click.option(
        '-m', '--model', type=click.Choice([m.value for m in Model], case_sensitive=False),
        default=Model.gain_blk_offset.value, show_default=True, help="""Correction model.
        \b
        
        - `gain`: Gain-only model.
        - `gain-blk-offset`: Gain-only model applied to offset normalised image blocks.
        - `gain-offset`: Gain and offset model.
        """,
    ),  # yapf: disable
    click.option(
        '-k', '--kernel-shape', type=click.Tuple([click.INT, click.INT]), nargs=2, default=(5, 5), show_default=True,
        metavar='HEIGHT WIDTH', help='Kernel height and width in pixels of the :option:`--proc-crs` image.'
    ),
    click.option(
        '-od', '--out-dir', type=click.Path(exists=True, file_okay=False, writable=True),
        show_default='source image directory.', help='Directory in which to place corrected image(s).'
    ),
    click.option(
        '-o', '--overwrite', is_flag=True, type=bool, default=False, show_default=True,
        help='Overwrite existing output file(s).'
    ),
    click.option(
        '-cmp', '--compare', 'comp_ref_file', metavar='FILE', type=click.Path(dir_okay=False, path_type=pathlib.Path),
        is_flag=False, flag_value='ref', default=None, callback=_compare_cb,
        help='Compare source and corrected images with this reference image.  If no ``FILE`` value is given, source '
        'and corrected images are compared with :option:`REFERENCE`.'
    ),
    click.option(
        '-bo/-nbo', '--build-ovw/--no-build-ovw', type=click.BOOL, default=True, show_default=True,
        help='Build overviews for the output image(s).'
    ),
    click.option(
        '-c', '--conf', type=click.Path(exists=True, dir_okay=False, readable=True, path_type=pathlib.Path),
        required=False, default=None, show_default=True,
        help='Path to a yaml configuration file specifying advanced options (as follow below).'
    )
)
# advanced options
@cloup.option_group(
    "Advanced options",
    click.option(
        '-pi/-npi', '--param-image/--no-param-image', type=click.BOOL, default=False, show_default=True,
        help=f'Write the  model parameters and R\N{SUPERSCRIPT TWO} values for each corrected image into a parameter '
        f'image file.'
    ),
    click.option(
        '-mp/-nmp', '--mask-partial/--no-mask-partial', type=click.BOOL,
        default=RasterFuse.create_model_config()['mask_partial'], show_default=True,
        help=f'Mask output pixels produced from partial kernel or source / reference image coverage.'
    ),
    # yapf: disable
    threads_option, max_block_mem_option, downsampling_option, upsampling_option,
    # yapf: enable
    click.option(
        '-rit', '--r2-inpaint-thresh', type=click.FloatRange(min=0, max=1),
        default=RasterFuse.create_model_config()['r2_inpaint_thresh'], show_default=True, metavar='FLOAT 0-1',
        help='R\N{SUPERSCRIPT TWO} threshold below which to inpaint model parameters from surrounding areas '
        '(0 = turn off inpainting). Valid for `gain-offset` :option:`--model` only.'
    ),
    click.option(
        '-pc', '--proc-crs', type=click.Choice([pc.value for pc in ProcCrs], case_sensitive=False),
        default=ProcCrs.auto.value, show_default=True,
        help="""The image CRS in which to estimate correction parameters. 
        \b
    
        - `auto`: lowest resolution of the source and reference CRS's (recommended). 
        - `src`: source image CRS. 
        - `ref`: reference image CRS.
        """
    ),
    click.option(
        '--driver',
        type=click.Choice(tuple(set(rio.drivers.raster_driver_extensions().values())), case_sensitive=False),
        default=RasterFuse.create_out_profile()['driver'], show_default=True, metavar='TEXT',
        help='Output image format driver.  See the `GDAL docs <https://gdal.org/drivers/raster/index.html>`_ for '
        'details.'
    ),  # yapf: disable
    click.option(
        '--dtype', type=click.Choice(list(rio.dtypes.dtype_fwd.values())[1:8], case_sensitive=False),
        default=RasterFuse.create_out_profile()['dtype'], show_default=True,
        help=f'Output image data type.  Valid for corrected images only, parameter images always use '
        f'{RasterArray.default_dtype}.'
    ),
    click.option(
        '--nodata', 'nodata', type=click.STRING, callback=_nodata_cb, metavar='[NUMBER|null|nan]',
        default=RasterFuse.create_out_profile()['nodata'], show_default=True,
        help=f'Output image nodata value.  Valid for corrected images only, parameter images always use '
        f'{RasterArray.default_nodata}.'
    ),
    click.option(
        '-co', '--creation-options', metavar='NAME=VALUE', multiple=True, default=(), callback=_creation_options_cb,
        help='Driver specific image creation option(s) for the output image(s).  See the `GDAL docs '
        '<https://gdal.org/drivers/raster/index.html>`_ for details.'
    ),
)
@click.pass_context
def fuse(
    ctx: click.Context, src_file: Tuple[pathlib.Path, ...], ref_file: pathlib.Path, model: Model,
    kernel_shape: Tuple[int, int], out_dir: pathlib.Path, overwrite: bool, comp_ref_file: pathlib.Path, build_ovw: bool,
    proc_crs: ProcCrs, conf: pathlib.Path, param_image: bool, **kwargs
):
    # @formatter:off
    """
    Correct image(s) to surface reflectance.

    Correct source multi-spectral aerial or satellite imagery to approximate surface reflectance, by fusion with a
    reference satellite image.

    For best results, reference and source image(s) should be concurrent, co-located, and spectrally similar.
    Reference image extents must encompass those of the source image(s), and source / reference band ordering should
    match i.e. reference band 1 corresponds to source band 1, reference band 2 corresponds to source band 2 etc.

    Corrected image(s) are named automatically based on the source file name and option values.
    \b

    Examples:
    ---------

    Correct `source.tif` with `reference.tif` using the default options::

        homonim fuse source.tif reference.tif

    Correct `source.tif` with `reference.tif` using the `gain-blk-offset` model, a kernel of 5 x 5 pixels,
    and placing the corrected images in the `corrected` directory::

        homonim fuse --model gain-blk-offset --kernel-shape 5 5 --out-dir ./corrected source.tif reference.tif

    Correct files matching `source*.tif` with `reference1.tif` using the `gain-offset` model and a kernel of 15 x 15
    pixels.  Produce parameter images, mask partially covered pixels in the corrected images, and statistically
    compare source and corrected images with `reference2.tif`::

        homonim fuse -m gain-offset -k 15 15 --param-image --mask-partial --compare reference2.tif source*.tif reference1.tif
    """
    # @formatter:on

    try:
        kernel_shape = utils.validate_kernel_shape(kernel_shape, model=model)
    except Exception as ex:
        raise click.BadParameter(str(ex))

    # build configuration dictionaries for RasterFuse
    block_config = _update_existing_keys(RasterFuse.create_block_config(), **kwargs)
    model_config = _update_existing_keys(RasterFuse.create_model_config(), **kwargs)
    out_profile = _update_existing_keys(RasterFuse.create_out_profile(), **kwargs)
    config = dict(block_config=block_config, model_config=model_config, out_profile=out_profile)
    comp_files = []

    # iterate over and correct source file(s)
    try:
        for src_filename in src_file:
            out_path = pathlib.Path(out_dir) if out_dir is not None else src_filename.parent
            logger.info(f'\nCorrecting {src_filename.name}')
            with RasterFuse(src_filename, ref_file, proc_crs=proc_crs) as raster_fuse:
                # construct output filenames
                post_fix = utils.create_out_postfix(
                    raster_fuse.proc_crs, model=model, kernel_shape=kernel_shape, driver=out_profile['driver'],
                )
                corr_filename = out_path.joinpath(src_filename.stem + post_fix)
                param_filename = utils.create_param_filename(corr_filename) if param_image else None

                start_time = timer()
                raster_fuse.process(
                    corr_filename, Model(model), kernel_shape, param_filename=param_filename, build_ovw=build_ovw,
                    overwrite=overwrite, **config,
                )

            logger.info(f'Completed in {timer() - start_time:.2f} secs')
            comp_files += [src_filename, corr_filename]  # build a list of files to pass to compare

        # compare source and corrected files with reference (invokes compare command with relevant parameters)
        if comp_ref_file:
            comp_file = ref_file if str(comp_ref_file) == 'ref' else comp_ref_file
            comp_cfg = {k: kwargs[k] for k, v in RasterCompare.create_config().items() if k in kwargs}
            ctx.invoke(compare, src_file=comp_files, ref_file=comp_file, proc_crs=proc_crs, **comp_cfg)

    except Exception:
        logger.exception('Exception caught during processing.')  # log exception info
        raise click.Abort()


cli.add_command(fuse)


# compare command
@cloup.command(cls=HomonimCommand)
@cloup.argument(
    'src-file', nargs=-1, metavar='IMAGE...', type=click.Path(exists=True, dir_okay=False, path_type=pathlib.Path),
    help='Path(s) to image(s) to compare with :option:`REFERENCE`.'
)
@ref_file_arg
@cloup.option_group("Standard options", output_option, )
@cloup.option_group(
    "Advanced options", threads_option, max_block_mem_option, downsampling_option, upsampling_option,
    click.option(
        '-pc', '--proc-crs', type=click.Choice([pc.value for pc in ProcCrs], case_sensitive=False),
        default=ProcCrs.auto.value, show_default=True, help="""The image CRS in which to compare images. 
    \b

    - `auto`: lowest resolution of the source and reference CRS's (recommended). 
    - `src`: source image CRS. 
    - `ref`: reference image CRS.
    """
    ),
)
def compare(
    src_file: Tuple[pathlib.Path, ...], ref_file: pathlib.Path, output: pathlib.Path, proc_crs: ProcCrs, **kwargs
):
    """
    Compare image(s) with a reference.

    Report similarity statistics between input image(s) and a reference image.  Typically, this is used to compare the
    before and after accuracy of surface reflectance correction, by comparing source and corrected images with a
    new reference image.

    Reference image extents must encompass those of the input image(s), and input / reference band ordering should
    match i.e. reference band 1 corresponds to input band 1, reference band 2 corresponds to input band 2 etc.
    \b

    Examples:
    ---------

    Compare `source.tif` and `corrected.tif` with `reference.tif`::

        homonim compare source.tif corrected.tif reference.tif
    """

    # build configuration dictionary
    config = RasterCompare.create_config(**kwargs)
    try:
        stats_dict = {}
        # iterate over source files, comparing with reference
        for src_filename in src_file:
            logger.info(f'\nComparing {src_filename.name}')
            start_time = timer()
            with RasterCompare(src_filename, ref_file, proc_crs=proc_crs) as raster_compare:
                stats_dict[str(src_filename)] = raster_compare.compare(**config)
            logger.info(f'Completed in {timer() - start_time:.2f} secs')

        # print a key for the following tables
        logger.info(f'\n\n{raster_compare.schema_table}')

        # print a results table per source image file
        summ_list = []
        for src_filename, im_stats_list in stats_dict.items():
            logger.info(f'\n\n{str(src_filename)}:\n\n{raster_compare.stats_table(im_stats_list)}')
            mean_dict = im_stats_list[-1].copy()
            mean_dict.pop('band')
            summ_list.append(dict(file=pathlib.Path(src_filename).name, **mean_dict))

        # print a summary results table comparing all source files
        if len(summ_list) > 1:
            logger.info(f'\n\nSummary over bands:\n\n{raster_compare.stats_table(summ_list)}')

        if output is not None:
            stats_dict['Reference'] = str(ref_file)
            with open(output, 'w') as file:
                json.dump(stats_dict, file)

    except Exception:
        logger.exception('Exception caught during processing')
        raise click.Abort()


cli.add_command(compare)


@cloup.command(cls=HomonimCommand)
@cloup.argument(
    'param-files', nargs=-1, metavar='PARAM...', type=click.Path(exists=True, dir_okay=False, path_type=pathlib.Path),
    callback=_param_file_cb, help='Path(s) to parameter image(s).'
)
@output_option
def stats(param_files: Tuple[pathlib.Path, ...], output: pathlib.Path):
    """
    Report parameter statistics.

    Report the minimum, maximum, mean etc. values of a parameter image generated with the
    :option:`--param-image <homonim-fuse --param-image>` option of the ``fuse`` command.
    """

    try:
        stats_dict = {}
        meta_dict = {}

        # process parameter file(s), storing results
        for param_filename in param_files:
            logger.info(f'\nProcessing {param_filename.name}')
            with ParamStats(param_filename) as param_stats:
                stats_dict[str(param_filename)] = param_stats.stats()
                meta_dict[str(param_filename)] = param_stats.metadata

        # print a key for the following tables
        logger.info(f'\n\n{param_stats.schema_table}')

        # iterate over stored result(s) and print
        for param_filename, param_dict in stats_dict.items():
            logger.info(f'\n{pathlib.Path(param_filename).name}:\n')
            logger.info(meta_dict[param_filename])
            logger.info(f'Stats:\n\n{ParamStats.stats_table(stats_dict[param_filename])}\n')

        if output is not None:
            with open(output, 'w') as file:
                json.dump(stats_dict, file, allow_nan=True)

    except Exception:
        logger.exception('Exception caught during processing')
        raise click.Abort()


cli.add_command(stats)

##



"""
    Homonim: Correction of aerial and satellite imagery to surface reflectance
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


class HomonimError(Exception):
    """ Root exception class. """


class UnsupportedImageError(HomonimError):
    """ Raised when reading a 12bit jpeg compressed image. """


class ImageContentError(HomonimError):
    """ Raised when an image has insufficient coverage or bands. """


class BlockSizeError(HomonimError):
    """ Raised when the image block size is invalid. """


class ImageProfileError(HomonimError):
    """ Raised when an image profile is invalid. """


class ImageFormatError(HomonimError):
    """ Raised when an image format is invalid. """


class IoError(HomonimError):
    """ Raised when accessing unopened file(s). """


class BandCountMismatchWarning(UserWarning):
    """ Warn when the reference and source band counts don't match. """


class NodataMaskWarning(UserWarning):
    """ Warn when an image has no mask or nodata values. """



"""
    Homonim: Correction of aerial and satellite imagery to surface reflectance
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from enum import Enum


class Model(str, Enum):
    """ Surface reflectance correction model. """
    gain = 'gain'
    """ 
    Gain-only model, suitable for haze-free and zero offset images (i.e. images where a surface reflectance of 
    zero corresponds to a pixel value of ~zero). 
    """
    gain_blk_offset = 'gain-blk-offset'
    """ 
    Gain-only model applied to offset normalised image blocks.  Suitable for most source / reference combinations.
    """
    gain_offset = 'gain-offset'
    """
    Gain and offset model.  The most accurate model, but sensitive to differences between source and reference, 
    such as shadowing and land cover changes.  Suitable for well-matched source / reference image pairs.  
    """


class ProcCrs(str, Enum):
    """
    CRS in which images will be processed.
    """
    auto = 'auto'
    """ Lowest resolution of the source and reference image CRS's (recommended). """
    src = 'src'
    """ Source image CRS. """
    ref = 'ref'
    """ Reference image CRS. """



"""
    Homonim: Correction of aerial and satellite imagery to surface reflectance
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import logging
import threading
from concurrent import futures
from contextlib import contextmanager
from pathlib import Path
from typing import Tuple, Dict, Union, Iterator, Optional

import numpy as np
import rasterio
import rasterio as rio
from rasterio.enums import Resampling
from rasterio.io import DatasetWriter
from tqdm import tqdm

from homonim import utils
from homonim.enums import Model, ProcCrs
from homonim.errors import IoError
from homonim.kernel_model import KernelModel, RefSpaceModel, SrcSpaceModel
from homonim.raster_array import RasterArray
from homonim.raster_pair import RasterPairReader, BlockPair

logger = logging.getLogger(__name__)


class RasterFuse(RasterPairReader):

    def __init__(
        self, src_filename: Union[Path, str], ref_filename: Union[Path, str], proc_crs: ProcCrs = ProcCrs.auto,
    ):
        """
        Class for correcting an image to surface reflectance, by fusion with a reference.

        Parameters
        ----------
        src_filename: str, pathlib.Path
            Path to the source image file.
        ref_filename: str, pathlib.Path
            Path to the reference image file.  The extents of this image should cover the source with at least a 2
            pixel border.  The reference image should have at least as many bands as the source, and the
            ordering of the source and reference bands should match.
        proc_crs: homonim.enums.ProcCrs, optional
            :class:`~homonim.enums.ProcCrs` instance specifying which of the source/reference image spaces should be
            used for estimating correction parameters.  In most cases, it can be left as the default of
            :attr:`~homonim.enums.ProcCrs.auto`,  where it will be resolved to the lowest resolution of the source and
            reference image CRS's.
        """
        RasterPairReader.__init__(self, src_filename, ref_filename, proc_crs=proc_crs)
        self._corr_lock = threading.Lock()
        self._param_lock = threading.Lock()

    create_model_config = KernelModel.create_config

    @staticmethod
    def create_block_config(threads: int = 0, max_block_mem: float = 100) -> Dict:
        """
        Utility method to create a block processing configuration dictionary that can be passed to
        :meth:`RasterFuse.process`.  Without arguments, the default configuration is returned.

        Parameters
        ----------
        threads: int, optional
            Number of image blocks to process concurrently.  A maximum of the number of processors on your
            system is allowed.  Increasing this number will increase the memory required for processing.
            0 = use all processors.
        max_block_mem: float, optional
            Maximum size of an image block in megabytes. Note that the total memory consumed by a thread is
            proportional to, but a number of times larger than this number.

        Returns
        -------
        dict
            Block processing configuration dictionary.
        """
        return dict(threads=utils.validate_threads(threads), max_block_mem=max_block_mem)

    @staticmethod
    def create_out_profile(
        driver: str = 'GTiff',
        dtype: str = RasterArray.default_dtype,
        nodata: float = RasterArray.default_nodata,
        creation_options: Optional[Dict] = None
    ) -> Dict:  # yapf: disable
        """
        Utility method to create a `rasterio` image profile for the output image(s) that can be passed to
        :meth:`RasterFuse.process`.  Without arguments, the default profile is returned.

        Parameters
        ----------
        driver: str, optional
            Output format driver.  See the `GDAL docs <https://gdal.org/drivers/raster/index.html>`_ for
            available options.
        dtype: str, optional
            Output image data type.  One of: uint8|uint16|int16|uint32|int32|float32|float64.
        nodata: float, optional
            Output image nodata value.
        creation_options: dict, optional
            Driver specific creation options e.g. ``dict(compression='deflate')`` for a GeoTIFF.
            See the `GDAL docs <https://gdal.org/drivers/raster/index.html>`_ for available keys and values.

        Returns
        -------
        dict
            `rasterio` image profile for output images.
        """
        # TODO: test effect of photometric=None on full size NGI files with ProcCrs.src param im (is bigtiff
        #  necessary then?).  Also test ovw compress with / w/o compress_overview
        creation_options = creation_options or dict(
            tiled=True, blockxsize=512, blockysize=512, compress='deflate', interleave='band', photometric=None,
            bigtiff='if_safer', compress_overview='auto',
        )
        return dict(driver=driver, dtype=dtype, nodata=nodata, creation_options=creation_options)

    @staticmethod
    def _build_overviews(im: DatasetWriter, max_num_levels: int = 8, min_level_pixels: int = 256):
        """
        Build internal overviews, downsampled by successive powers of 2, for an open rasterio dataset.
        Overviews are limited so that the highest level has at least ``min_level_pixels`` pixels along the shortest
        dimension, and so there are no more than ``max_num_levels`` levels.
        """
        if im.closed:
            raise IoError(f'The raster dataset is closed: {im.name}')

        max_ovw_levels = int(np.min(np.log2(im.shape)))
        min_level_shape_pow2 = int(np.log2(min_level_pixels))
        num_ovw_levels = np.min([max_num_levels, max_ovw_levels - min_level_shape_pow2])
        ovw_levels = [2 ** m for m in range(1, num_ovw_levels + 1)]
        im.build_overviews(ovw_levels, Resampling.average)

    def _merge_corr_profile(self, out_profile: Optional[Dict] = None) -> Dict:
        """
        Return a rasterio profile for the corrected image, by merging the source image profile with ``out_profile``.
        """
        out_profile = self.create_out_profile(**(out_profile or {}))
        corr_profile = utils.combine_profiles(self.src_im.profile, out_profile)
        corr_profile['count'] = len(self.src_bands)
        return corr_profile

    def _merge_param_profile(self, out_profile: Optional[Dict] = None) -> Dict:
        """
        Create a rasterio profile for the parameter image, using a merge of the ``proc_crs`` image profile,
        and ``out_profile`` as a starting point.
        """
        if self.proc_crs == ProcCrs.ref:
            init_profile = self.ref_im.profile
        else:
            init_profile = self.src_im.profile
        out_profile = self.create_out_profile(**(out_profile or {}))
        param_profile = utils.combine_profiles(init_profile, out_profile)
        # force dtype and nodata to defaults
        param_profile.update(
            dtype=RasterArray.default_dtype, count=len(self.src_bands) * 3, nodata=RasterArray.default_nodata
        )
        return param_profile

    def _set_metadata(self, im: DatasetWriter, **kwargs):
        """
        Utility method to convert ``**kwargs``, and RasterPairReader attributes to configuration metadata in an
        open rasterio dataset.
        """
        if im.closed:
            raise IoError(f'The raster dataset is closed: {im.name}')

        kwargs_meta_dict = {f'FUSE_{k.upper()}': v.name if hasattr(v, 'name') else v for k, v in kwargs.items()}
        meta_dict = dict(
            FUSE_SRC_FILE=self._src_filename.name, FUSE_REF_FILE=self._ref_filename.name,
            FUSE_PROC_CRS=self.proc_crs.name, **kwargs_meta_dict,
        )
        im.update_tags(**meta_dict)

    def _set_corr_metadata(self, im: DatasetWriter, **kwargs):
        """
        Utility method to convert ``**kwargs`` configuration and reference band info to metadata in a corrected
        image.
        """
        if im.closed:
            raise IoError(f'The raster dataset is closed: {im.name}')

        self._set_metadata(im, **kwargs)
        for bi in range(0, min(im.count, len(self.ref_bands))):
            ref_bi = self.ref_bands[bi]
            ref_meta_dict = self.ref_im.tags(ref_bi)
            # TODO: update to copy the latest geedim metadata (also update the test data) (makes sense to do this
            #  with the band matching update)
            corr_meta_dict = {k: v for k, v in ref_meta_dict.items() if k in ['ABBREV', 'ID', 'NAME']}
            im.set_band_description(bi + 1, self.ref_im.descriptions[ref_bi - 1])
            im.update_tags(bi + 1, **corr_meta_dict)

    def _set_param_metadata(self, im: DatasetWriter, **kwargs):
        """
        Utility method to convert ``**kwargs`` to configuration metadata ino a parameter image, and set the band
        metadata to describe the corresponding parameter.
        """
        if im.closed:
            raise IoError(f'The raster dataset is closed: {im.name}')

        self._set_metadata(im, **kwargs)
        num_src_bands = len(self.src_bands)
        for bi in range(0, num_src_bands):
            ref_bi = self.ref_bands[bi]
            ref_descr = self.ref_im.descriptions[ref_bi - 1] or f'B{ref_bi}'
            ref_meta_dict = self.ref_im.tags(ref_bi)
            param_names = ['GAIN', 'OFFSET', 'R2']
            for param_i, param_name in zip(range(bi, im.count, num_src_bands), param_names):
                im.set_band_description(param_i + 1, f'{ref_descr}_{param_name}')
                param_meta_dict = {
                    k: f'{v.upper()} {param_name}' for k, v in ref_meta_dict.items() if k in ['ABBREV', 'ID', 'NAME']
                }
                im.update_tags(param_i + 1, **param_meta_dict)

    @contextmanager
    def _out_files(
        self, corr_filename: Union[Path, str], param_filename: Union[Path, str] = None, out_profile: Dict = None,
        overwrite: bool = False, build_ovw: bool = False, **kwargs
    ) -> Iterator[Tuple[rasterio.DatasetReader, Union[rasterio.DatasetReader, None]]]:
        """
        Internal context manager to handle the corrected, and optional parameter, output file(s).

        On entry, the image files are configured and created using ``out_profile``.
        On exit, image metadata is set with ``**kwargs``, overviews are built if ``build_ovw`` is True, and the file(s)
        are closed.

        Existing files are not overwritten unless ``overwrite`` is True.
        """
        # entry
        if not overwrite and corr_filename.exists():
            raise FileExistsError(
                f"Corrected image file exists and won't be overwritten without the `overwrite` option: {corr_filename}"
            )
        if not overwrite and param_filename and param_filename.exists():
            raise FileExistsError(
                f"Parameter image file exists and won't be overwritten without the `overwrite` option: {param_filename}"
            )
        out_im = rio.open(corr_filename, 'w', **self._merge_corr_profile(out_profile))
        param_im = rio.open(param_filename, 'w', **self._merge_param_profile(out_profile)) if param_filename else None
        try:
            yield out_im, param_im
        finally:
            # exit
            self._set_corr_metadata(out_im, **kwargs)
            if build_ovw:
                self._build_overviews(out_im)
            out_im.close()
            if param_im:
                self._set_param_metadata(param_im, **kwargs)
                if build_ovw:
                    self._build_overviews(param_im)
                param_im.close()

    def _process_block(
        self, block_pair: BlockPair, model: KernelModel, corr_im: DatasetWriter,
        param_im: Optional[DatasetWriter] = None,
    ):
        """
        Thread-safe method to correct an image block to surface reflectance using the supplied correction ``model``.
        Corrected, and optionally parameter, blocks are written to the supplied image dataset(s).
        """
        # read source and reference blocks
        src_ra, ref_ra = self.read(block_pair)
        # fit and apply the sliding kernel models
        param_ra = model.fit(src_ra, ref_ra)
        corr_ra = model.apply(src_ra, param_ra)
        # change the corrected nodata value so that is masked correctly for corr_im
        corr_ra.nodata = corr_im.nodata

        with self._corr_lock:  # write the corrected block
            corr_ra.to_rio_dataset(corr_im, indexes=block_pair.band_i + 1, window=block_pair.src_out_block)

        if param_im:
            with self._param_lock:  # write the parameter block
                indexes = np.arange(param_ra.count) * len(self.src_bands) + block_pair.band_i + 1
                param_out_block = (
                    block_pair.ref_out_block if self.proc_crs == ProcCrs.ref else block_pair.src_out_block
                )
                param_ra.to_rio_dataset(param_im, indexes=indexes, window=param_out_block)

    def process(
        self,
        corr_filename: Union[Path, str],
        model: Model,
        kernel_shape: Tuple[int, int],
        param_filename: Optional[Union[Path, str]] = None,
        build_ovw: bool = True,
        overwrite: bool = False,
        model_config: Optional[Dict] = None,
        out_profile: Optional[Dict] = None,
        block_config: Optional[Dict] = None,
    ):  # yapf: disable
        r"""
        Correct the source image to surface reflectance by fusion with the reference.

        To improve speed and reduce memory usage, images are divided into blocks for concurrent processing.

        Parameters
        ----------
        corr_filename: str, Path
            Path to the corrected file to create.
        model: homonim.enums.Model
            The surface reflectance correction model to use.  See :class:`~homonim.enums.Model` for details.
        kernel_shape: tuple of int
            The (height, width) of the kernel in pixels of the :attr:`proc_crs` image (the lowest resolution
            image, if :attr:`proc_crs` is :attr:`~homonim.enums.ProcCrs.auto`).
        param_filename: str, Path, optional
            Path to an optional parameter file to write with correction parameters and *R*\ :sup:`2` values.  By
            default, no parameter file is written.
        build_ovw: bool, optional
            Build overviews for the output image(s).
        overwrite: bool, optional
            Overwrite the output image(s) if they exist.
        model_config: dict, optional
            Configuration dictionary for the correction model.  See
            :meth:`create_model_config` for possible keys and default values.
        out_profile: dict, optional
            Configuration dictionary for the output image(s).   See :meth:`~RasterFuse.create_out_profile` for
            possible keys and default values.
        block_config: dict, optional
            Configuration dictionary for block processing.  See :meth:`~RasterFuse.create_block_config` for possible
            keys and default values.
        """
        self._assert_open()

        # prepare configuration
        model_type = Model(model)
        kernel_shape = tuple(utils.validate_kernel_shape(kernel_shape, model=model))
        overlap = utils.overlap_for_kernel(kernel_shape)
        model_config = RasterFuse.create_model_config(**(model_config or {}))
        block_config = RasterFuse.create_block_config(**(block_config or {}))

        # create the KernelModel according to proc_crs
        model_cls = SrcSpaceModel if self.proc_crs == ProcCrs.src else RefSpaceModel
        model = model_cls(model, kernel_shape, find_r2=param_filename is not None, **model_config)

        # arguments to self.block_pairs()
        block_pair_args = dict(overlap=overlap, max_block_mem=block_config['max_block_mem'])
        # tqdm progress bar format
        bar_format = '{l_bar}{bar}|{n_fmt}/{total_fmt} blocks [{elapsed}<{remaining}]'

        # create and open the output files
        with self._out_files(
            corr_filename, param_filename=param_filename, out_profile=out_profile, overwrite=overwrite,
            build_ovw=build_ovw, model=model_type, kernel_shape=kernel_shape, **model_config, **block_config
        ) as (out_im, param_im):  # yapf: disable
            if block_config['threads'] == 1:
                # correct blocks consecutively in the main thread (useful for profiling)
                block_pairs = [block_pair for block_pair in self.block_pairs(**block_pair_args)]
                for block_pair in tqdm(block_pairs, bar_format=bar_format):
                    self._process_block(block_pair, model, corr_im=out_im, param_im=param_im)
            else:
                # correct blocks concurrently
                with futures.ThreadPoolExecutor(max_workers=block_config['threads']) as executor:
                    # submit block correction jobs to the thread pool
                    proc_futures = [
                        executor.submit(self._process_block, block_pair, model, out_im, param_im)
                        for block_pair in self.block_pairs(**block_pair_args)
                    ]

                    # wait for threads in order of completion, and raise any thread generated exceptions
                    for future in tqdm(
                        futures.as_completed(proc_futures), bar_format=bar_format, total=len(proc_futures),
                        dynamic_ncols=True,
                    ):  # yapf: disable
                        future.result()



"""
    Homonim: Correction of aerial and satellite imagery to surface reflectance
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import logging
import pathlib
import threading
from contextlib import ExitStack
from itertools import product
from typing import Tuple, NamedTuple, Union, Iterable

import numpy as np
import rasterio
import rasterio as rio
from rasterio.enums import MaskFlags
from rasterio.vrt import WarpedVRT
from rasterio.warp import Resampling
from rasterio.windows import Window
from tqdm.contrib.logging import logging_redirect_tqdm

from homonim import errors, utils
from homonim.enums import ProcCrs
from homonim.raster_array import RasterArray

logger = logging.getLogger(__name__)


class BlockPair(NamedTuple):
    """ A set of matching block windows for a source-reference image pair. """
    band_i: int
    """ Band index (0 based). """
    src_in_block: Window
    """ Overlapping / input source window. """
    ref_in_block: Window
    """ Overlapping / input reference window. """
    src_out_block: Window
    """ Non-overlapping / output source window. """
    ref_out_block: Window
    """ Non-overlapping / output reference window. """
    outer: bool
    """ True if any part of the source blocks touch the source image boundary, otherwise False. """


class RasterPairReader:

    def __init__(
        self, src_filename: Union[str, pathlib.Path], ref_filename: Union[str, pathlib.Path],
        proc_crs: ProcCrs = ProcCrs.auto
    ):
        """
        Class for reading matching, and optionally overlapping, blocks from a source and reference image pair.

        Parameters
        ----------
        src_filename: str, pathlib.Path
            Path to the source image file.
        ref_filename: str, pathlib.Path
            Path to the reference image file.  The extents of this image should cover the source with at least a 2
            pixel border.  The reference image should have at least as many bands as the source, and the
            ordering of the source and reference bands should match.
        proc_crs: homonim.enums.ProcCrs, optional
            :class:`~homonim.enums.ProcCrs` instance specifying which of the source/reference image spaces will be
            used for processing.  For most use cases, including typical surface reflectance correction,
            it can be left as the default of :attr:`~homonim.enums.ProcCrs.auto`. In this case it will be resolved to
            refer to the lowest resolution of the source and reference image CRS's.
        """
        self._src_filename = pathlib.Path(src_filename)
        self._ref_filename = pathlib.Path(ref_filename)
        proc_crs = ProcCrs(proc_crs)
        self._proc_crs = proc_crs

        self._env = None
        self._src_lock = threading.Lock()
        self._ref_lock = threading.Lock()
        self._src_im = None
        self._ref_im = None
        self._stack = None
        self._init_image_pair()

    @property
    def src_im(self) -> rasterio.DatasetReader:
        """ Source rasterio dataset. """
        self._assert_open()
        return self._src_im

    @property
    def ref_im(self) -> rasterio.DatasetReader:
        """ Reference rasterio dataset. """
        self._assert_open()
        return self._ref_im

    @property
    def src_bands(self) -> Tuple[int, ...]:
        """ Source non-alpha band indices (1-based). """
        return self._src_bands

    @property
    def ref_bands(self) -> Tuple[int, ...]:
        """ Reference non-alpha band indices (1-based). """
        return self._ref_bands

    @property
    def proc_crs(self) -> ProcCrs:
        """ Which of the source and reference image CRS's will be used for processing. """
        return self._proc_crs

    @property
    def closed(self) -> bool:
        """ True if both source and reference images are closed, otherwise False. """
        return not self._src_im or not self._ref_im or self._src_im.closed or self._ref_im.closed

    @staticmethod
    def _validate_image(im: rasterio.DatasetReader):
        """ Validate an open rasterio dataset for use as a source or reference image. """

        try:
            _ = im.read(1, window=im.block_window(1, 0, 0))
        except Exception as ex:
            if 'compress' in im.profile and im.profile['compress'] == 'jpeg':  # assume it is a 12bit JPEG
                raise errors.UnsupportedImageError(
                    f'Could not read image {im.name}.  JPEG compression with NBITS==12 is not supported, '
                    f'you probably need to recompress this image.'
                )
            else:
                raise ex

        # warn if there is no nodata or associated mask
        is_masked = any([MaskFlags.all_valid not in im.mask_flag_enums[bi] for bi in range(im.count)])
        if im.nodata is None and not is_masked:
            logger.warning(
                f'{im.name} has no mask or nodata value, any invalid pixels should be masked before processing.'
            )

    @staticmethod
    def _validate_image_pair(src_im: rasterio.DatasetReader, ref_im: rasterio.DatasetReader):
        """ Validate a pair of rasterio datasets for use as a source-reference image pair. """

        for im in (src_im, ref_im):
            RasterPairReader._validate_image(im)
        # check reference image extents cover the source
        if not utils.covers_bounds(ref_im, src_im):
            raise errors.ImageContentError(f'Reference extent does not cover source image')

        # retrieve non-alpha bands
        src_bands = utils.get_nonalpha_bands(src_im)
        ref_bands = utils.get_nonalpha_bands(ref_im)

        # check reference has enough bands
        if len(src_bands) > len(ref_bands):
            raise errors.ImageContentError(
                f'Reference ({ref_im.name}) has fewer non-alpha bands than source ({src_im.name}).'
            )

        # warn if source and reference band counts don't match
        if len(src_bands) != len(ref_bands):
            logger.warning(
                f'Source and reference image non-alpha band counts don`t match. Using the first {len(src_bands)} '
                f'non-alpha bands of reference.'
            )

        # warn if the source and reference are not in the same CRS
        if src_im.crs.to_proj4() != ref_im.crs.to_proj4():
            logger.warning(f'Source and reference image pair are not in the same CRS: {src_im.name} and {ref_im.name}')

    @staticmethod
    def _resolve_proc_crs(
        src_im: rasterio.DatasetReader, ref_im: rasterio.DatasetReader, proc_crs: ProcCrs = ProcCrs.auto
    ) -> ProcCrs:
        """
        Resolve a :class:`~homonim.enums.ProcCrs` instance.
        The :class:`~homonim.enums.ProcCrs` instance is resolved from :attr:`~homonim.enums.ProcCrs.auto` to the lowest
        resolution CRS of the supplied source and reference images.  If the :class:`~homonim.enums.ProcCrs` instance is
        already resolved, and doesn't correspond to the lowest resolution CRS, then a warning is issued.
        """
        # compare source and reference resolutions
        src_pixel_smaller = np.prod(np.abs(src_im.res)) <= np.prod(np.abs(ref_im.res))
        cmp_str = 'smaller' if src_pixel_smaller else 'larger'
        if proc_crs == ProcCrs.auto:
            # set proc_crs to the lowest resolution of the source and reference images
            proc_crs = ProcCrs.ref if src_pixel_smaller else ProcCrs.src
            logger.debug(
                f'Source pixel size {np.round(src_im.res, decimals=3)} is {cmp_str} than the reference '
                f'{np.round(ref_im.res, decimals=3)}. Using proc_crs=`{proc_crs}`.'
            )
        elif (
            (proc_crs == ProcCrs.src and src_pixel_smaller) or
            (proc_crs == ProcCrs.ref and not src_pixel_smaller)
        ):  # yapf: disable
            # warn if the proc_crs value does not correspond to the lowest resolution of the source and
            # reference images
            rec_crs_str = ProcCrs.ref if src_pixel_smaller else ProcCrs.src
            logger.warning(
                f'proc_crs={rec_crs_str} is recommended when the source pixel size is {cmp_str} than the reference.'
            )
        return proc_crs

    @staticmethod
    def resolve_proc_crs(
        src_filename: pathlib.Path, ref_filename: pathlib.Path, proc_crs: ProcCrs = ProcCrs.auto
    ) -> ProcCrs:
        """
        Resolve a :class:`~homonim.enums.ProcCrs` instance.

        The :class:`~homonim.enums.ProcCrs` instance is resolved from :attr:`~homonim.enums.ProcCrs.auto` to the lowest
        resolution CRS of the supplied source and reference images.  If the :class:`~homonim.enums.ProcCrs` instance is
        already resolved, and doesn't correspond to the lowest resolution CRS, then a warning is issued.

        Parameters
        ----------
        src_filename: str, pathlib.Path
            Path to the source image file.
        ref_filename: str, pathlib.Path
            Path to the reference image file.
        proc_crs: homonim.enums.ProcCrs, optional
            :class:`~homonim.enums.ProcCrs` instance to resolve.

        Returns
        -------
        ProcCrs
            :class:`~homonim.enums.ProcCrs` instance resolved to either :attr:`~homonim.enums.ProcCrs.src` or
            :attr:`~homonim.enums.ProcCrs.ref`.
        """
        with rio.open(src_filename, 'r') as src_im, rio.open(ref_filename, 'r') as _ref_im:
            with (
                WarpedVRT(_ref_im, crs=src_im.crs, resampling=Resampling.bilinear)
                if src_im.crs.to_proj4() != _ref_im.crs.to_proj4() else _ref_im
            ) as ref_im:  # yapf: disable
                return RasterPairReader._resolve_proc_crs(src_im, ref_im, proc_crs=proc_crs)

    def _auto_block_shape(self, max_block_mem: float = np.inf) -> Tuple[int, int]:
        """ Find a block shape that satisfies max_block_mem. """

        proc_win = self._ref_win if self.proc_crs == ProcCrs.ref else self._src_win
        # adjust max_block_mem to represent the size of a block in the highest resolution image, but scaled to the
        # equivalent in proc_crs.
        src_pix_area = np.product(np.abs(self._src_im.res))
        ref_pix_area = np.product(np.abs(self._ref_im.res))
        if self.proc_crs == ProcCrs.ref:
            mem_scale = src_pix_area / ref_pix_area if ref_pix_area > src_pix_area else 1.
        elif self.proc_crs == ProcCrs.src:
            mem_scale = 1. if ref_pix_area > src_pix_area else ref_pix_area / src_pix_area
        else:
            raise ValueError("'proc_crs' has not been resolved - the raster pair must be opened first.")
        max_block_mem = max_block_mem * mem_scale if max_block_mem > 0 else np.inf

        max_block_mem *= 2 ** 20  # convert MB to bytes
        dtype_size = np.dtype(RasterArray.default_dtype).itemsize  # the size of the RasterArray data type

        # set the starting block_shape to correspond to the entire window
        block_shape = np.array((proc_win.height, proc_win.width)).astype('float')

        # keep halving the block_shape along the longest dimension until it satisfies max_block_mem
        while (np.product(block_shape) * dtype_size) > max_block_mem:
            div_dim = np.argmax(block_shape)
            block_shape[div_dim] /= 2

        if np.any(block_shape < (1, 1)):
            raise errors.BlockSizeError(f"The auto block shape is smaller than a pixel.  Increase 'max_block_mem'.")

        block_shape = np.ceil(block_shape).astype('int')
        logger.debug(
            f'Auto block shape: {block_shape}, of image shape: {[proc_win.height, proc_win.width]}'
            f' ({self.proc_crs.name} pixels)'
        )

        # warn if the block shape in the highest res image is less than a typical tile
        if np.any(block_shape / mem_scale < (256, 256)) and np.any(block_shape < (proc_win.height, proc_win.width)):
            logger.warning(
                f'The auto block shape is small: {block_shape}.  Increase `max_block_mem` to improve processing times.'
            )
        return tuple(block_shape)

    def _init_image_pair(self):
        """ Prepare the raster pair for reading. """
        # check and resolve proc_crs before opening
        self._proc_crs = self.resolve_proc_crs(self._src_filename, self._ref_filename, proc_crs=self._proc_crs)
        self.open()
        try:
            self._validate_image_pair(self._src_im, self._ref_im)
            # get non-alpha band indices for reading
            self._src_bands = utils.get_nonalpha_bands(self._src_im)
            logger.debug(f'{self._src_filename.name} non-alpha bands: {self._src_bands}')
            self._ref_bands = utils.get_nonalpha_bands(self._ref_im)
            logger.debug(f'{self._ref_filename.name} non-alpha bands: {self._ref_bands}')

            # create image windows that allow re-projections between source and reference without loss of data
            self._ref_win = utils.expand_window_to_grid(self._ref_im.window(*self._src_im.bounds))
            self._src_win = utils.expand_window_to_grid(self._src_im.window(*self._ref_im.window_bounds(self._ref_win)))

        finally:
            self.close()

    def _assert_open(self):
        """ Raise an IoError if the source and reference images are not open. """
        if self.closed:
            raise errors.IoError(
                f'The raster pair has not been opened: {self._src_filename.name} and {self._ref_filename.name}'
            )

    def open(self):
        """ Open the source and reference images for reading. """
        self._src_im = rio.open(self._src_filename, 'r')
        self._ref_im = rio.open(self._ref_filename, 'r')

        # It is essential that the source and reference are in the same CRS so that rectangular regions of valid
        # data in one will re-project to rectangular regions of valid data in the other.
        if self._src_im.crs.to_proj4() != self._ref_im.crs.to_proj4():
            # open the image pair in the same CRS, re-projecting the proc_crs (usually lower resolution) image into the
            # CRS of the other
            if self.proc_crs == ProcCrs.src:
                self._src_im = WarpedVRT(self._src_im, crs=self._ref_im.crs, resampling=Resampling.bilinear)
            else:
                self._ref_im = WarpedVRT(self._ref_im, crs=self._src_im.crs, resampling=Resampling.bilinear)

    def close(self):
        """ Close the source and reference image datasets. """
        self._src_im.close()
        self._ref_im.close()

    def __enter__(self):
        self._stack = ExitStack()
        self._stack.enter_context(rio.Env(GDAL_NUM_THREADS='ALL_CPUs'))
        self._stack.enter_context(logging_redirect_tqdm([logging.getLogger(__package__)]))
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        self._stack.__exit__(exc_type, exc_val, exc_tb)

    def read(self, block_pair: BlockPair) -> Tuple[RasterArray, RasterArray]:
        """
        Thread-safe read of a matching pair of source and reference image blocks.

        Parameters
        ----------
        block_pair: BlockPair
            :class:`BlockPair` named tuple, specifying the :attr:`BlockPair.src_in_block` source window, and
            :attr:`BlockPair.ref_in_block` reference window to read.  This :class:`BlockPair` instance is
            typically obtained from :meth:`block_pairs`.

        Returns
        -------
        src_ra: RasterArray
            Source image block wrapped in a RasterArray.
        ref_ra: RasterArray
            Reference image block wrapped in a RasterArray.
        """
        self._assert_open()

        with self._src_lock:
            src_ra = RasterArray.from_rio_dataset(
                self._src_im, indexes=self._src_bands[block_pair.band_i], window=block_pair.src_in_block
            )
        with self._ref_lock:
            ref_ra = RasterArray.from_rio_dataset(
                self._ref_im, indexes=self._ref_bands[block_pair.band_i], window=block_pair.ref_in_block
            )
        return src_ra, ref_ra

    def block_pairs(self, overlap: Tuple[int, int] = (0, 0), max_block_mem: float = np.inf) -> Iterable[BlockPair]:
        """
        Iterator over paired source-reference image blocks.

        Parameters
        ----------
        overlap: tuple of int
            Block overlap (rows, columns) in pixels of the :attr:`proc_crs` image.
        max_block_mem: float
            Maximum allowable block size in MB. The image is divided into 2\\ :sup:`n` blocks with n the smallest number
            where ``max_block_mem`` is satisfied.  If ``max_block_mem`` is `float('inf')`, the block shape will be set
            to encompass the full extent of the source image.

        Yields
        -------
        BlockPair
            Named tuple specifying the overlapping (``*_in_block``), and non-overlapping (``*_out_block``) image blocks.
        """
        self._assert_open()
        # generate the auto block shape for reading
        overlap = np.array(overlap).astype('int')
        block_shape = self._auto_block_shape(max_block_mem=max_block_mem)
        if np.any(block_shape <= overlap):
            raise errors.BlockSizeError(f'The auto block shape is smaller than the overlap.  Increase `max_block_mem`.')
        logger.debug(f'Block overlap: {overlap} ({self.proc_crs.name} pixels)')

        # initialise block formation variables
        # blocks are first formed in proc_crs, then transformed to the 'other' image crs, so here we assign the src/ref
        # windows etc. to proc_* equivalents
        if self.proc_crs == ProcCrs.ref:
            proc_win, proc_im, other_im = (self._ref_win, self._ref_im, self._src_im)
        else:
            proc_win, proc_im, other_im = (self._src_win, self._src_im, self._ref_im)

        proc_win_ul = np.array((proc_win.row_off, proc_win.col_off))
        proc_win_br = np.array((proc_win.height + proc_win.row_off, proc_win.width + proc_win.col_off))

        # Outer loop over bands so that all blocks in a band are yielded consecutively - this is fastest for
        # reading band interleaved images.
        for band_i in range(len(self._src_bands)):
            # Inner loop over the upper left corner row, col for each overlapping block
            ul_row_range = range(
                proc_win.row_off - overlap[0], proc_win.row_off + proc_win.height - overlap[0], block_shape[0]
            )
            ul_col_range = range(
                proc_win.col_off - overlap[1], proc_win.col_off + proc_win.width - overlap[1], block_shape[1]
            )
            for ul_row, ul_col in product(ul_row_range, ul_col_range):
                # find UL and BR corners for overlapping block in proc space
                ul = np.array((ul_row, ul_col))
                br = ul + block_shape + (2 * overlap)
                # limit block extents to image window extents
                in_ul = np.fmax(ul, proc_win_ul)
                in_br = np.fmin(br, proc_win_br)
                # find UL and BR corners for non-overlapping block in proc space
                out_ul = np.fmax(ul + overlap, proc_win_ul)
                out_br = np.fmin(br - overlap, proc_win_br)
                # block touches image boundary?
                outer = np.any(in_ul <= proc_win_ul) or np.any(in_br >= proc_win_br)

                # Create rasterio windows corresponding to above block corners.
                # Note:
                # - Consecutive proc_in_block's will overlap by exactly ``overlap``.
                # - Consecutive proc_out_block's will be exactly adjacent.
                proc_in_block = Window(*in_ul[::-1], *np.subtract(in_br, in_ul)[::-1])
                proc_out_block = Window(*out_ul[::-1], *np.subtract(out_br, out_ul)[::-1])

                # Create equivalent rasterio windows in 'other' space.
                # Note:
                # - other_in_block boundaries are expanded to ensure that re-projecting between source/reference
                #   CRSs does not mask valid *_out_block pixels.  This means that consecutive other_in_blocks may
                #   overlap by more than ``overlap``.
                # - consecutive other_out_block's may overlap by a pixel.
                other_in_block = utils.expand_window_to_grid(other_im.window(*proc_im.window_bounds(proc_in_block)))
                other_out_block = utils.round_window_to_grid(other_im.window(*proc_im.window_bounds(proc_out_block)))

                # create the BlockPair named tuple, assigning 'proc' and 'other' back to 'src' and 'ref' for passing to
                # read()
                if self.proc_crs == ProcCrs.ref:
                    block_pair = BlockPair(
                        band_i, other_in_block, proc_in_block, other_out_block, proc_out_block, outer
                    )
                else:
                    block_pair = BlockPair(
                        band_i, proc_in_block, other_in_block, proc_out_block, other_out_block, outer
                    )
                yield block_pair



"""
    Homonim: Correction of aerial and satellite imagery to surface reflectance
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Tuple, Dict, Optional

import cv2 as cv
import numpy as np
from rasterio.enums import Resampling
from rasterio.fill import fillnodata

from homonim import utils
from homonim.enums import Model
from homonim.raster_array import RasterArray

ONdArray = Optional[np.ndarray]
OShape = Optional[Tuple[int, int]]


class KernelModel:

    def __init__(self, model: Model, kernel_shape: Tuple[int, int], find_r2: bool = False, **kwargs):
        """
        A base class for surface reflectance modelling and correction of blocks of image data.

        The surface reflectance relationship between source and reference image blocks is approximated with localised
        linear models.  Models are estimated for each pixel location inside a small rectangular kernel (window),
        using a fast DFT approach.  The corrected output is produced by applying the model parameters to the source
        image block.

        Based on the paper:
        Harris, Dugal & Van Niekerk, Adriaan. (2018). Radiometric homogenisation of aerial images by calibrating with
        satellite data. International Journal of Remote Sensing. 40. 1-25. 10.1080/01431161.2018.1528404.
        https://www.researchgate.net/publication
        /328317307_Radiometric_homogenisation_of_aerial_images_by_calibrating_with_satellite_data

        Parameters
        ----------
        model: homonim.enums.Model
            Surface reflectance correction model.
        kernel_shape: tuple
            (height, width) of the kernel in pixels.
        find_r2: bool, optional
            Whether to calculate *R*\\ :sup:`2` (coefficient of determination) for each kernel model, and include in
            parameter arrays.
        kwargs:
            Optional configuration arguments.  See :meth:`KernelModel.create_config` for keys and defaults values.
        """
        self._model = Model(model)
        self._kernel_shape = utils.validate_kernel_shape(kernel_shape, model=model)
        self._find_r2 = find_r2

        # update config defaults with any passed values, and set attributes
        config = self.create_config(**kwargs)
        self._r2_inpaint_thresh: float = config['r2_inpaint_thresh']
        self._mask_partial: bool = config['mask_partial']
        self._downsampling: Resampling = config['downsampling']
        self._upsampling: Resampling = config['upsampling']

    @property
    def model(self) -> Model:
        """ Correction model type. """
        return self._model

    @property
    def kernel_shape(self) -> Tuple[int, int]:
        """ Kernel (height, width) in pixels. """
        return tuple(self._kernel_shape)

    @property
    def find_r2(self) -> bool:
        """ Whether to create and include R2 (coefficient of determination) in parameter arrays. """
        return self._find_r2

    @staticmethod
    def create_config(
        r2_inpaint_thresh: float = 0.25, mask_partial: bool = False, downsampling: Resampling = Resampling.average,
        upsampling: Resampling = Resampling.cubic_spline,
    ) -> Dict:
        """
        Utility method to create a KernelModel configuration dictionary that can be passed to
        :meth:`KernelModel.__init__` and :meth:`homonim.RasterFuse.process`.  Without arguments, the default
        configuration values are returned.

        Parameters
        ----------
        r2_inpaint_thresh: float, optional
            *R*\\ :sup:`2` (coefficient of determination) threshold below which to `in-paint` kernel model parameters
            from surrounding areas (applies to the :attr:`~homonim.enums.Model.gain_offset` model only).  For
            pixels where the model gives a poor approximation to the data (this can occur in areas where source and
            reference differ due to e.g. shadowing, land cover changes etc.), model offsets are interpolated from
            surrounding areas, and gains re-estimated.  Set ``r2_inpaint_thresh`` to `None` to turn off in-painting.
        mask_partial: bool, optional
            Mask output pixels not produced by full kernel or source/reference image coverage.  Useful for ensuring
            strict model validity, and reducing seam-lines between overlapping images.
        downsampling: rasterio.enums.Resampling, optional
            Resampling method to use when downsampling.  See the `rasterio docs
            <https://rasterio.readthedocs.io/en/latest/api/rasterio.enums.html#rasterio.enums.Resampling>`_ for
            available options.
        upsampling: rasterio.enums.Resampling, optional
            Resampling method to use when upsampling. See the `rasterio docs
            <https://rasterio.readthedocs.io/en/latest/api/rasterio.enums.html#rasterio.enums.Resampling>`_ for
            available options.

        Returns
        -------
        dict
            Configuration dictionary.
        """
        return dict(
            r2_inpaint_thresh=r2_inpaint_thresh, mask_partial=mask_partial, downsampling=downsampling,
            upsampling=upsampling,
        )

    def _get_resampling(self, from_res: Tuple[float, float], to_res: Tuple[float, float]):
        """ Return the resampling method for re-projecting from resolution ``from_res`` to resolution ``to_res``. """
        return self._downsampling if np.prod(np.abs(from_res)) <= np.prod(np.abs(to_res)) else self._upsampling

    def _r2_array(
        self, ref_array: np.ndarray, src_array: np.ndarray, param_array: np.ndarray, mask: ONdArray = None,
        mask_sum: ONdArray = None, ref_sum: ONdArray = None, src_sum: ONdArray = None, ref2_sum: ONdArray = None,
        src2_sum: ONdArray = None, src_ref_sum: ONdArray = None, dest_array: ONdArray = None,
        kernel_shape: OShape = None
    ) -> np.ndarray:
        """
        Utility function to return R2 (coefficient of determination) at each pixel/kernel location for the given
        arrays.
        """
        if kernel_shape is None:
            kernel_shape = self._kernel_shape
        kernel_shape = tuple(kernel_shape)  # force to tuple for opencv
        filter_args = dict(normalize=False, borderType=cv.BORDER_CONSTANT)  # common opencv arguments

        # find the keyword arguments that were not provided
        if mask is None:
            # if mask is passed, assume that it has been applied to ref_array and src_array, otherwise do that here
            mask = (
                ~utils.nan_equals(src_array, RasterArray.default_nodata) &
                ~utils.nan_equals(ref_array, RasterArray.default_nodata)
            )
            ref_array[~mask] = 0
            src_array[~mask] = 0
        if mask_sum is None:
            mask_sum = cv.boxFilter(mask.astype(RasterArray.default_dtype), -1, kernel_shape[::-1], **filter_args)
        if ref_sum is None:
            ref_sum = cv.boxFilter(ref_array, -1, kernel_shape[::-1], **filter_args)
        if ref2_sum is None:
            ref2_sum = cv.sqrBoxFilter(ref_array, -1, kernel_shape[::-1], **filter_args)
        if src2_sum is None:
            src2_sum = cv.sqrBoxFilter(src_array, -1, kernel_shape[::-1], **filter_args)
        if src_ref_sum is None:
            src_ref_sum = cv.boxFilter(src_array * ref_array, -1, kernel_shape[::-1], **filter_args)

        # R2 is found using: R2 = 1 - (residual sum of squares)/(total sum of squares) = 1 - RSS/TSS
        # TSS = sum((ref - mean(ref))**2), which can be expanded and expressed in terms of cv.boxFilter kernel sums as:
        ss_tot_array = (mask_sum * ref2_sum) - (ref_sum ** 2)

        if param_array.shape[0] > 1:
            # find RSS for model == Model.gain_offset
            if src_sum is None:
                src_sum = cv.boxFilter(src_array, -1, kernel_shape[::-1], **filter_args)

            # RSS = sum((ref - ref_hat)**2)
            #     = sum((ref - (m*src + c))**2), where m and c are the first 2 bands of param_array
            # The above can be expanded and expressed in terms of cv.boxFilter kernel sums as:
            ss_res_array = (
                ((param_array[0] ** 2) * src2_sum) +
                (2 * np.product(param_array[:2], axis=0) * src_sum) -
                (2 * param_array[0] * src_ref_sum) -
                (2 * param_array[1] * ref_sum) +
                ref2_sum + (mask_sum * (param_array[1] ** 2))
            )  # yapf: disable
        else:
            # find RSS for model == Model.gain or Model.gain_blk_offset

            # RSS = sum((ref - m*src)**2), where m is the first band of param_array
            # The above can be expanded and expressed in terms of cv.boxFilter kernel sums as:
            ss_res_array = (((param_array[0] ** 2) * src2_sum) - (2 * param_array[0] * src_ref_sum) + ref2_sum)

        ss_res_array *= mask_sum

        if dest_array is None:
            # assign a destination array to write R2 into, if it was not provided
            dest_array = np.full(
                src_array.shape, fill_value=RasterArray.default_nodata, dtype=RasterArray.default_dtype
            )

        # find R2 = 1 - RSS/TSS, and write into dest_array
        np.divide(ss_res_array, ss_tot_array, out=dest_array, where=mask)
        np.subtract(1, dest_array, out=dest_array, where=mask)
        return dest_array

    @staticmethod
    def _fit_block_norm(src_ra: RasterArray, ref_ra: RasterArray) -> np.ndarray:
        """
        Return a two element [gain, offset] model to 'normalise' the source block, so that the standard deviation and
        first percentile of the source and reference blocks match.  (Can be thought of as a basic dark object
        subtraction (DOS) approach).
        """
        norm_model = np.zeros(2)
        mask = ref_ra.mask & src_ra.mask
        if not np.any(mask):
            return norm_model
        norm_model[0] = np.std(ref_ra.array[mask]) / np.std(src_ra.array[mask])
        norm_model[1] = np.percentile(ref_ra.array[mask], 1) - np.percentile(src_ra.array[mask], 1) * norm_model[0]
        return norm_model

    def _fit_gain(self, src_ra: RasterArray, ref_ra: RasterArray, kernel_shape: OShape = None) -> RasterArray:
        """
        Find sliding kernel gains, for a source & reference band, using opencv convolution.
        Returns a RasterArray of sliding kernel model parameters. Gains in first band, offsets in the second,
        and optionally R2 for each kernel model in the third band when :attr:`find_r2` is True.
        """
        # adapted from https://www.mathsisfun.com/data/least-squares-regression.html with c=0
        if kernel_shape is None:
            kernel_shape = self._kernel_shape
        kernel_shape = tuple(kernel_shape)  # force to tuple for opencv

        # mask invalid pixels with 0 so that these do not contribute to kernel sums in *boxFilter()
        ref_array = ref_ra.array
        src_array = src_ra.array
        mask = ref_ra.mask & src_ra.mask
        ref_array[~mask] = 0
        src_array[~mask] = 0

        # set up a RasterArray profile for the parameters
        param_profile = src_ra.profile.copy()
        param_profile.update(
            count=3 if self._find_r2 else 2, nodata=RasterArray.default_nodata, dtype=RasterArray.default_dtype
        )

        # convolve the kernel with src and ref to get kernel sums (uses DFT for large kernels)
        filter_args = dict(normalize=False, borderType=cv.BORDER_CONSTANT)  # common opencv arguments
        src_sum = cv.boxFilter(src_array, -1, kernel_shape[::-1], **filter_args)
        ref_sum = cv.boxFilter(ref_array, -1, kernel_shape[::-1], **filter_args)

        # create parameter RasterArray filled with nodata
        param_ra = RasterArray.from_profile(None, param_profile)
        param_ra.array[1, mask] = 0  # set offsets to 0

        # find sliding kernel gains, avoiding divide by 0
        np.divide(ref_sum, src_sum, out=param_ra.array[0], where=mask)

        if self._find_r2:
            # Find R2 of the sliding kernel models
            self._r2_array(
                ref_array, src_array, param_ra.array[:1], mask=mask, ref_sum=ref_sum, src_sum=src_sum,
                dest_array=param_ra.array[2], kernel_shape=kernel_shape
            )

        return param_ra

    def _fit_gain_blk_offset(
        self, src_ra: RasterArray, ref_ra: RasterArray, kernel_shape: OShape = None
    ) -> RasterArray:  # yapf: disable
        """
        Find sliding kernel gains and 'image' (i.e. block) offset, for a band, using opencv convolution.
        Returns a RasterArray of sliding kernel model parameters. Gains in first band, offsets in the second,
        and optionally R2 for each kernel model in the third band when :attr:`find_r2` is True.
        """
        if kernel_shape is None:
            kernel_shape = self._kernel_shape
        kernel_shape = tuple(kernel_shape)  # force to tuple for opencv

        # find the source->reference normalisation
        norm_model = self._fit_block_norm(src_ra, ref_ra)

        # force src nodata to nan so that operation below remains correctly masked
        src_ra.nodata = RasterArray.default_nodata

        # apply the normalisation (block gain and offset)
        src_ra.array = (src_ra.array * norm_model[0]) + norm_model[1]

        # find gains for normalised source
        param_ra = self._fit_gain(src_ra, ref_ra, kernel_shape=kernel_shape)

        # incorporate the normalisation model in the parameter RasterArray
        param_ra.array[1] = param_ra.array[0] * norm_model[1]
        param_ra.array[0] *= norm_model[0]
        return param_ra

    def _fit_gain_offset(self, src_ra: RasterArray, ref_ra: RasterArray, kernel_shape: OShape = None) -> RasterArray:
        """
        Find sliding kernel full linear model for a band using opencv convolution.
        Returns a RasterArray of sliding kernel model parameters. Gains in first band, offsets in the second,
        and optionally R2 for each kernel model in the third band when :attr:`find_r2` is True.
        """
        # Least squares formulae adapted from https://www.mathsisfun.com/data/least-squares-regression.html
        if kernel_shape is None:
            kernel_shape = self._kernel_shape
        kernel_shape = tuple(kernel_shape)  # force to tuple for opencv

        # mask invalid pixels with 0 so that these do not contribute to kernel sums in *boxFilter()
        ref_array = ref_ra.array
        src_array = src_ra.array
        mask = ref_ra.mask & src_ra.mask
        ref_array[~mask] = 0
        src_array[~mask] = 0

        # set up a RasterArray profile for the parameters
        param_profile = src_ra.profile.copy()
        find_r2 = self._find_r2 or (self._r2_inpaint_thresh is not None)
        param_profile.update(
            count=3 if find_r2 else 2, nodata=RasterArray.default_nodata, dtype=RasterArray.default_dtype
        )

        # find the numerator for the gain i.e N*cov(ref, src)
        filter_args = dict(normalize=False, borderType=cv.BORDER_CONSTANT)  # common opencv arguments
        src_sum = cv.boxFilter(src_array, -1, kernel_shape[::-1], **filter_args)
        ref_sum = cv.boxFilter(ref_array, -1, kernel_shape[::-1], **filter_args)
        src_ref_sum = cv.boxFilter(src_array * ref_array, -1, kernel_shape[::-1], **filter_args)
        mask_sum = cv.boxFilter(
            mask.astype(RasterArray.default_dtype, copy=False), -1, kernel_shape[::-1], **filter_args
        )
        m_num_array = (mask_sum * src_ref_sum) - (src_sum * ref_sum)

        # find the denominator for the gain i.e. N*var(src)
        src2_sum = cv.sqrBoxFilter(src_array, -1, kernel_shape[::-1], **filter_args)
        m_den_array = (mask_sum * src2_sum) - (src_sum ** 2)

        # create parameter RasterArray filled with nodata
        param_ra = RasterArray.from_profile(None, param_profile)

        # find the gain = cov(ref, src) / var(src), avoiding divide by 0
        np.divide(m_num_array, m_den_array, out=param_ra.array[0], where=mask)

        # solve for the offset c = y - mx, given that the linear model passes through (mean(ref), mean(src))
        np.divide(ref_sum - (param_ra.array[0] * src_sum), mask_sum, out=param_ra.array[1], where=mask)

        if find_r2:
            # Find R2 of the sliding kernel models
            self._r2_array(
                ref_array, src_array, param_ra.array[:2], mask=mask, mask_sum=mask_sum, ref_sum=ref_sum,
                src_sum=src_sum, src2_sum=src2_sum, src_ref_sum=src_ref_sum, dest_array=param_ra.array[2],
                kernel_shape=kernel_shape
            )

        if self._r2_inpaint_thresh is not None:
            # fill/inpaint low R2 and negative gain areas in the offset parameter
            r2_mask = (param_ra.array[2] > self._r2_inpaint_thresh) & (param_ra.array[0] > 0) & mask
            param_ra.array[1] = fillnodata(param_ra.array[1], r2_mask)
            param_ra.mask = mask  # re-mask as nodata areas will have been filled above

            # recalculate the gain for the filled areas using m = (y - c)/x and and the point (mean(ref), mean(src)))
            r2_mask = ~r2_mask & mask
            np.divide(ref_sum - mask_sum * param_ra.array[1], src_sum, out=param_ra.array[0], where=r2_mask)

        return param_ra

    def _full_coverage_mask(self, in_mask_ra: RasterArray, param_ra: RasterArray) -> RasterArray:
        """
        Utility function to create a full coverage mask i.e. a mask of output pixels fully covered by source/reference
        data, sliding (model) kernels, and resampling kernels.

        Note: this is a strict approach that avoids any kind of partial coverage.

        Parameters
        ----------
        in_mask_ra: RasterArray
            Initial mask of valid source/reference pixels, as a RasterArray with dtype=='uint8'.
        param_ra: RasterArray
            A parameter RasterArray as returned by :meth:`fit` i.e. a RasterArray in the CRS and grid corresponding
            to :attr:`proc_crs`, whose mask corresponds to the combined reference & source masks.

        Returns
        -------
        RasterArray
            Full coverage mask as a RasterArray with dtype=='uint8'
        """

        # re-project the initial mask into proc_crs (the CRS and grid corresponding to the proc_crs attribute)
        mask_ra = in_mask_ra.reproject(**param_ra.proj_profile, nodata=None, resampling=Resampling.average)
        # find the mask of proc_crs pixels fully covered by mask_ra
        mask = (mask_ra.array >= 1).astype('uint8', copy=False)  # ref pixels fully covered by src
        # combine mask with the other (src/ref) mask via param_ra
        mask &= param_ra.mask

        # Mask out partial kernel coverage.
        # Similar to the block overlap amount, this removes ceil(kernel_shape/2) pixels from the nodata edge.  Note,
        # that this is the strict approach for proc_crs == ref, it could be floor(kernel_shape/2) for proc_crs == src,
        # which avoids the additional upsampling step.
        se = cv.getStructuringElement(cv.MORPH_RECT, tuple(np.array(self._kernel_shape[::-1]) + 2))
        mask_ra.array = cv.erode(mask, se, borderType=cv.BORDER_CONSTANT, borderValue=0)
        return mask_ra

    def fit(self, src_ra: RasterArray, ref_ra: RasterArray) -> RasterArray:
        """
        Fit sliding kernel models to reference and source blocks.

        Parameters
        ----------
        src_ra: RasterArray
            Source data block in a RasterArray, with the same CRS, shape & extents as ``ref_ra``.
        ref_ra: RasterArray
            Reference data block in a RasterArray.

        Returns
        -------
        RasterArray
            RasterArray of sliding kernel model parameters. Gains in first band, offsets in the second, and optionally
            *R*\\ :sup:`2` for each kernel model in the third band when :attr:`find_r2` is True.
        """
        # TODO : include a CRS comparison below i.e. one that is faster that rasterio's current implementation
        if (ref_ra.transform != src_ra.transform) or (ref_ra.shape != src_ra.shape):
            raise ValueError("'ref_ra' and 'src_ra' must have the same CRS, transform and shape")

        if self._model == Model.gain:
            param_ra = self._fit_gain(src_ra, ref_ra, kernel_shape=self._kernel_shape)
        elif self._model == Model.gain_blk_offset:
            param_ra = self._fit_gain_blk_offset(src_ra, ref_ra, kernel_shape=self._kernel_shape)
        else:
            param_ra = self._fit_gain_offset(src_ra, ref_ra, kernel_shape=self._kernel_shape)

        return param_ra

    def apply(self, src_ra: RasterArray, param_ra: RasterArray) -> RasterArray:
        """
        Apply kernel models to a source block.

        Parameters
        ----------
        src_ra: RasterArray
            Source data block in a RasterArray.
        param_ra: RasterArray
            RasterArray of sliding kernel model parameters. Gains in first band, offsets in the second, and the same
            CRS, shape & extents as src_ra.

        Returns
        -------
        corr_ra :RasterArray
            Corrected block in a RasterArray.
        """
        if (param_ra.transform != src_ra.transform) or (param_ra.shape != src_ra.shape):
            raise ValueError("'param_ra' and 'src_ra' must have the same CRS, transform and shape")
        corr_array = (param_ra.array[0] * src_ra.array) + param_ra.array[1]
        corr_ra = RasterArray.from_profile(corr_array, param_ra.profile)
        return corr_ra


class RefSpaceModel(KernelModel):
    """
    A KernelModel subclass, for estimating model parameters in the reference image CRS and grid.

    The source image block is re-projected into the reference CRS to estimate the parameters.  Estimated parameters are
    subsequently re-projected to the source CRS for application to the (original) source image block.

    Recommended for the most common use case where the reference image has a lower resolution than the source image.
    """

    def fit(self, src_ra: RasterArray, ref_ra: RasterArray) -> RasterArray:
        # choose resampling method based on whether we are up- or downsampling
        resampling = self._get_resampling(src_ra.res, ref_ra.res)
        # downsample src_ra to reference CRS and grid
        src_ds_ra = src_ra.reproject(**ref_ra.proj_profile, resampling=resampling)
        # call base class fit with reference and source RasterArrays in the reference CRS & grid
        return KernelModel.fit(self, src_ds_ra, ref_ra)

    def apply(self, src_ra, param_ra):
        # create a parameter RasterArray containing only the first two bands of param_ra
        # (to speed up the re-projection below)
        _param_ra = RasterArray.from_profile(param_ra.array[:2], param_ra.profile)
        # choose resampling method based on whether we are up- or downsampling
        resampling = self._get_resampling(param_ra.res, src_ra.res)
        # re-project param_ra to source CRS and grid
        param_us_ra = _param_ra.reproject(**src_ra.proj_profile, resampling=resampling)

        if self._mask_partial:
            # find the mask of fully covered pixels in reference CRS and grid
            mask_ra = self._full_coverage_mask(src_ra.mask_ra, _param_ra)
            # re-project the mask to source CRS and grid, and apply to the parameters
            mask_us_ra = mask_ra.reproject(**src_ra.proj_profile, nodata=0, resampling=Resampling.nearest)
            param_us_ra.mask = mask_us_ra.array.astype('bool', copy=False)
        else:
            param_us_ra.mask = src_ra.mask

        # call base class apply with source and parameter RasterArrays in the source CRS & grid
        return KernelModel.apply(self, src_ra, param_us_ra)


class SrcSpaceModel(KernelModel):
    """
    A KernelModel subclass, for estimating model parameters in the source image CRS and grid.

    The reference image block is re-projected into the source CRS to estimate the parameters.  Estimated parameters are
    subsequently applied (directly) to the source image block.

    Recommended for the unusual use case where the source image has a lower resolution than the reference image.
    """

    def fit(self, src_ra: RasterArray, ref_ra: RasterArray) -> RasterArray:
        # choose resampling method based on whether we are up- or downsampling
        resampling = self._get_resampling(ref_ra.res, src_ra.res)
        # upsample ref_ra to the source CRS and grid
        ref_us_ra = ref_ra.reproject(**src_ra.proj_profile, resampling=resampling)

        _src_ra = src_ra.copy()  # copy the source to avoid in-place changes in fit() below
        # call base class fit with source and parameter RasterArrays in the source CRS & grid
        param_ra = KernelModel.fit(self, _src_ra, ref_us_ra)

        if self._mask_partial:
            # remove r2 band from param masking
            _param_ra = RasterArray.from_profile(param_ra.array[:2], param_ra.profile)
            # find the mask of fully covered pixels in source CRS and grid, and apply it to the parameters
            mask_ra = self._full_coverage_mask(ref_ra.mask_ra, _param_ra)
            param_ra.mask = mask_ra.array.astype('bool', copy=False)
        else:
            param_ra.mask = src_ra.mask

        return param_ra



"""
    Homonim: Correction of aerial and satellite imagery to surface reflectance
    Copyright (C) 2021 Dugal Harris
    Email: dugalh@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import logging
import pathlib
import threading
from concurrent import futures
from contextlib import ExitStack
from multiprocessing import cpu_count
from typing import List, Dict, Tuple, Union, Optional

import numpy as np
import rasterio as rio
import yaml
from rasterio.windows import get_data_window, intersect, union, Window
from tabulate import tabulate
from tqdm import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from homonim import utils, errors
from homonim.enums import Model

logger = logging.getLogger(__name__)


class ParamStats:

    def __init__(self, param_filename: Union[pathlib.Path, str]):
        """
        Class to calculate the statistics of a parameter image.

        Parameters
        ----------
        param_filename: pathlib.Path, str
            Path to a parameter image file, as created by :meth:`homonim.RasterFuse.process` with a specified
            ``param_filename``.  See the :meth:`homonim.RasterFuse.process` documentation for more details.
        """
        self._param_filename = pathlib.Path(param_filename)

        utils.validate_param_image(self._param_filename)

        # read some parameters from the metadata
        self._param_im: rio.DatasetReader
        with rio.open(self._param_filename, 'r') as self._param_im:
            self._tags = self._param_im.tags()
            self._model = self._tags['FUSE_MODEL'].replace('_', '-')
            self._r2_inpaint_thresh = yaml.safe_load(self._tags['FUSE_R2_INPAINT_THRESH'])

    schema = dict(
        band=dict(abbrev='Band'),
        mean=dict(abbrev='Mean'),
        std=dict(abbrev='Std.'),
        min=dict(abbrev='Min.'),
        max=dict(abbrev='Max.'),
        inpaint_p=dict(abbrev='Inpaint (%)', description='Portion of inpainted pixels (%).'),
    )  # yapf: disable
    """ Dictionary describing the statistics returned by :attr:`ParamStats.stats`. """

    @property
    def closed(self) -> bool:
        """ True if the parameter file is closed, otherwise False. """
        return not self._param_im or self._param_im.closed

    @property
    def metadata(self) -> str:
        """ Parameter metadata string. """
        res_str = (
            f'Model: {self._model}\n'
            f'Kernel shape: {self._tags["FUSE_KERNEL_SHAPE"]}\n'
            f'Processing CRS: {self._tags["FUSE_PROC_CRS"]}\n'
            f'Reference: {self._tags["FUSE_REF_FILE"]}\n'
        )
        if self._model == 'gain-offset':
            res_str += f'R\N{SUPERSCRIPT TWO} inpaint threshold: {self._r2_inpaint_thresh}\n'
        return res_str

    @property
    def schema_table(self) -> str:
        """ Table string describing statistics returned by :attr:`ParamStats.stats`. """
        schema_list = [v for k, v in self.schema.items() if 'description' in v]
        schema_list.append(dict(abbrev='*_R2', description='R\N{SUPERSCRIPT TWO} coefficient of determination.'))
        headers = {k: k.upper() for k in schema_list[0].keys()}
        return tabulate(schema_list, headers=headers, tablefmt=utils.table_format)

    @staticmethod
    def stats_table(stats_list: List[Dict]) -> str:
        """
        Create a table string from the provided parameter statistics.

        Parameters
        ----------
        stats_list: list of dict
            Parameter statistics to tabulate, as returned by :meth:`ParamStats.stats`.

        Returns
        -------
        str
            A table string.
        """
        headers = {k: v['abbrev'] for k, v in ParamStats.schema.items() if k in stats_list[-1]}
        return tabulate(stats_list, headers=headers, floatfmt='.3f', stralign='right', tablefmt=utils.table_format)

    def __enter__(self):
        self._stack = ExitStack()
        self._stack.enter_context(rio.Env(GDAL_NUM_THREADS='ALL_CPUs'))
        self._stack.enter_context(logging_redirect_tqdm([logging.getLogger(__package__)]))
        self._param_im = rio.open(self._param_filename, 'r')
        self._stack.enter_context(self._param_im)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._stack.__exit__(exc_type, exc_val, exc_tb)

    def _assert_open(self):
        """ Utility method to raise an exception if the parameter file is not open. """
        if self.closed:
            raise errors.IoError(f'The parameter file has not been opened: {self._param_filename.name}')

    def _get_data_window(self, threads: int = cpu_count()) -> Window:
        """ Threaded function to accumulate the parameter image data window over blocks.  """
        self._assert_open()
        read_lock = threading.Lock()

        def get_block_data_window(block_win: Window) -> Optional[Window]:
            """ Return a window of valid data corresponding to a given block.  """
            with read_lock:
                mask = self._param_im.read_masks(indexes=1, window=block_win)
            _block_data_win = get_data_window(mask, nodata=0)

            if _block_data_win.width == 0 or _block_data_win.height == 0:
                # return None if there is no valid data in the block
                return None
            # offset _block_data_win to the UL corner of block_win and return
            return Window(
                block_win.col_off + _block_data_win.col_off, block_win.row_off + _block_data_win.row_off,
                _block_data_win.width, _block_data_win.height
            )

        # combine valid block windows into a valid image window
        im_data_win: Optional[Window] = None
        bar_format = 'Finding window: {l_bar}{bar}|{n_fmt}/{total_fmt} blocks [{elapsed}<{remaining}]'
        with futures.ThreadPoolExecutor(max_workers=threads) as executor:
            # create threads to get valid block windows
            win_futures = [
                executor.submit(get_block_data_window, block_win)
                for block_ij, block_win in self._param_im.block_windows(1)
            ]  # yapf: disable

            # wait for threads and combine returned windows
            for future in tqdm(
                futures.as_completed(win_futures), bar_format=bar_format, total=len(win_futures), dynamic_ncols=True,
                leave=False,
            ):  # yapf: disable
                block_data_win: Window = future.result()
                if block_data_win:
                    im_data_win = union(im_data_win, block_data_win) if im_data_win else block_data_win
        return im_data_win

    def _get_image_stats(self, image_accum: List[Dict]) -> List[Dict]:
        """ Utility method to calculate image statistics from accumulated results. """
        image_stats = []
        for band_i, band_accum in enumerate(image_accum):
            band_stats = dict(
                band=self._param_im.descriptions[band_i],
                mean=band_accum['sum'] / band_accum['n'],
                # formula for cumulative std dev from:
                # https://rosettacode.org/wiki/Cumulative_standard_deviation#Python
                std=np.sqrt((band_accum['sum2'] / band_accum['n']) - (band_accum['sum'] ** 2 / band_accum['n'] ** 2)),
                min=band_accum['min'],
                max=band_accum['max'],
            )  # yapf: disable

            if 'inpaint_sum' in band_accum:
                band_stats['inpaint_p'] = 100 * band_accum['inpaint_sum'] / band_accum['n']
            image_stats.append(band_stats)
        return image_stats

    def stats(self, threads: int = 0) -> List[Dict]:
        """
        Find parameter image statistics.

        Statistics are accumulated over image blocks, allowing large images to be described with limited memory use.

        Parameters
        ----------
        threads: int, optional
            Number of image blocks to process concurrently.  A maximum of the number of processors on your
            system is allowed.  Increasing this number will increase the memory required for processing.
            0 = use all processors.

        Returns
        -------
        list of dict
            A list of parameter band statistics.
        """
        self._assert_open()
        threads = utils.validate_threads(threads)
        data_win = self._get_data_window(threads=threads)  # get valid data window
        read_lock = threading.Lock()

        def get_block_sums(band_i: int, block_win: rio.windows.Window) -> Tuple[Dict, int]:
            """ Thread-safe function read a block and find its sums etc. """
            with read_lock:
                array: np.ma.masked_array = self._param_im.read(
                    indexes=band_i + 1, window=block_win, masked=True, out_dtype='float64',
                )
            _block_dict = dict(
                min=array.min(), max=array.max(), sum=array.sum(), sum2=(array ** 2).sum(), n=array.count()
            )
            if (self._model == Model.gain_offset) and (band_i >= self._param_im.count * 2 / 3):
                # find the sum of inpainted pixels, if this is a R2 band and a gain-offset model
                _block_dict.update(inpaint_sum=(array < self._r2_inpaint_thresh).sum())
            return _block_dict, band_i

        # find block sums, min, max etc. in threads, and accumulate over the image
        image_accum = [{} for _ in range(self._param_im.count)]
        bar_format = 'Finding stats: {l_bar}{bar}|{n_fmt}/{total_fmt} blocks [{elapsed}<{remaining}]'
        with futures.ThreadPoolExecutor(max_workers=threads) as executor:
            # create threads to get block sums etc.
            stats_futures = [
                executor.submit(get_block_sums, band_i, block_win)
                for band_i in range(self._param_im.count)
                for block_ij, block_win in self._param_im.block_windows(band_i + 1)
                if intersect(data_win, block_win)
            ]  # yapf: disable

            # wait for threads, and accumulate block results over the image
            for future in tqdm(
                futures.as_completed(stats_futures), bar_format=bar_format, total=len(stats_futures),
                dynamic_ncols=True,
            ):  # yapf: disable
                block_dict, band_i = future.result()
                image_accum[band_i].update(
                    min=np.nanmin((image_accum[band_i].get('min', np.inf), block_dict['min'])),
                    max=np.nanmax((image_accum[band_i].get('max', -np.inf), block_dict['max'])),
                    sum=np.nansum((image_accum[band_i].get('sum', 0), block_dict['sum'])),
                    sum2=np.nansum((image_accum[band_i].get('sum2', 0), block_dict['sum2'])),
                    n=np.nansum((image_accum[band_i].get('n', 0), block_dict['n']))
                )
                if 'inpaint_sum' in block_dict:
                    image_accum[band_i].update(
                        inpaint_sum=np.nansum((image_accum[band_i].get('inpaint_sum', 0), block_dict['inpaint_sum']))
                    )

        # find the image statistics and return
        return self._get_image_stats(image_accum)
